/* $Id: libwebfwlog_udf.c 696 2019-09-20 16:47:49Z bhockney $ */

/* (C) 2010-2019 by Bob Hockney <zeus@ix.netcom.com>    *
 *                                                      *
 * MySQL UDF for IPV6 functions                         *
 *                                                      *
 * This code is distributed under the terms of GNU GPL  */

#include <ctype.h>
#include <string.h>
#include <netdb.h>
#include <math.h>
#include <mysql.h>

#include <main.h>
#include <utils.h>

/* 4 and 16 byte address lengths */
#define INET_ADDRLEN (sizeof(struct in_addr))
#define INET6_ADDRLEN (sizeof(struct in6_addr))

#ifndef MIN
  #define MIN(a,b) (((a)<(b))?(a):(b))
#endif

extern struct options opt;

my_bool wfwl_inet_pton_init(UDF_INIT *initid, UDF_ARGS *args, char *message);
char *wfwl_inet_pton(UDF_INIT *initid, UDF_ARGS *args, char *result,
        unsigned long *length, char *null_value, char *error);

my_bool wfwl_inet_ntop_init(UDF_INIT *initid, UDF_ARGS *args, char *message);
char *wfwl_inet_ntop(UDF_INIT *initid, UDF_ARGS *args, char *result,
        unsigned long *length, char *null_value, char *error);
void wfwl_inet_ntop_deinit(UDF_INIT *initid);

my_bool wfwl_inet_contains_init(UDF_INIT *initid, UDF_ARGS *args, char *message);
long long wfwl_inet_contains(UDF_INIT *initid, UDF_ARGS *args, char *result,
        unsigned long *length, char *null_value, char *error);

/**
 * wfwl_inet_pton()
 *
 * Convert IPv4 or IPv6 presentation string to VARBINARY(16) representation.
 *
 * Example: SELECT wfwl_inet_pton('1.2.3.4'), wfwl_inet_pton('fe80::219:e3ff:1:9317');
 *
 * @arg1   integer  address family, AF_INET or AF_INET6
 * @arg2   string   human readable ipv4 or ipv6 address
 * @return string   4 or 16 byte binary string
 */
my_bool wfwl_inet_pton_init(UDF_INIT *initid, UDF_ARGS *args, char *message)
{
    if (args->arg_count != 2 || (args->args[0] != NULL && args->arg_type[0] != INT_RESULT)
                             || args->arg_type[1] != STRING_RESULT) {
        strcpy(message, "Bad arguments: wfwl_inet_pton(address family, human readable IPv4 or IPv6 address)");
        return 1;
    }
    initid->max_length = INET6_ADDRLEN;
    initid->maybe_null = 1;
    initid->const_item = 1;
    args->maybe_null[0] = 1;
    return 0;
}

char *wfwl_inet_pton(UDF_INIT *initid, UDF_ARGS *args, char *result,
        unsigned long *res_length, char *null_value, char *error)
{
    char addr[INET6_ADDRSTRLEN + 1];
    unsigned int length;
    int family;

    if (!args->args[1] || !(length = args->lengths[1])) {
        *null_value = 1;
        return 0;
    }

    /* cannot assume null-terminated string according to manual */
    if (length >= sizeof(addr))
        length = sizeof(addr) - 1;
    memcpy(addr, args->args[1], length);
    addr[length] = 0;

    /* address family */
    if(args->args[0] == NULL) {
      if (strpbrk(addr, ":")) {
          family = AF_INET6;
          length = INET6_ADDRLEN;
      } else {
          family = AF_INET;
          length = INET_ADDRLEN;
      }
    } else {
      if (*args->args[0] == AF_INET) {
        family = *args->args[0];
        length = INET_ADDRLEN;
      } else if (*args->args[0] == AF_INET6 || *args->args[0] == 10) {
        if (! strpbrk(addr, ":")) {
          strncpy(addr, "::ffff:", 8);
          memcpy(addr + 7, args->args[1], MIN(INET6_ADDRSTRLEN - 7,length));
          addr[MIN(INET6_ADDRSTRLEN, length + 7)] = 0;
        }
        family = AF_INET6;
        length = INET6_ADDRLEN;
      } else {
        *null_value = 1;
        return 0;
      }
    }

    /* convert */
    if (inet_pton(family, addr, result) != 1) {
        *null_value = 1;
        return 0;
    }
    *res_length = length;
    return result;
}

/**
 * wfwl_inet_ntop()
 *
 * Convert IPv4 or IPv6 VARBINARY(16) format to presentation string.
 *
 * Example: SELECT wfwl_inet_ntop(2, wfwl_inet_pton('1.2.3.4')), wfwl_inet_ntop(10, wfwl_inet_pton('fe80::219:e3ff:1:9317'));
 *
 * @arg1   integer  address family, AF_INET or AF_INET6
 * @arg2   string   varbinary format ipv4 or ipv6 address
 * @return string   max 46 character presentation string
 */

my_bool wfwl_inet_ntop_init(UDF_INIT *initid, UDF_ARGS *args, char *message)
{
    if (args->arg_count != 2 || (args->args[0] != NULL && args->arg_type[0] != INT_RESULT)
                             || args->arg_type[1] != STRING_RESULT || args->lengths[1] > INET6_ADDRLEN) {
        sprintf(message, "bad arguments: wfwl_inet_ntop(address family, 4 or 16 byte binary string)");
        return 1;
    }
    initid->ptr = malloc(INET6_ADDRSTRLEN + 1);
    if (initid->ptr == NULL) {
        sprintf(message, "wfwl_inet_ntop: out of memory");
        return 1;
    }

    initid->max_length = INET6_ADDRSTRLEN; /* max length of ipv6 presentation string */
    initid->maybe_null = 1;
    initid->const_item = 1;
    args->maybe_null[0] = 1;
    return 0;
}

char *wfwl_inet_ntop(UDF_INIT *initid, UDF_ARGS *args, char *result,
        unsigned long *res_length, char *null_value, char *error)
{
    char *addr = NULL;
    unsigned char family;
    int j;
    char in[16];
    char *out = initid->ptr;
    char *prefix = "::ffff:";

    if (args->args[1] == NULL) {
        *null_value = 1;
        return 0;
    }

    if (args->lengths[1])
        memcpy(&in, args->args[1], args->lengths[1]);
    
    if (args->lengths[1] > INET_ADDRLEN)
        for (j = INET6_ADDRLEN - 1; j >= (int)args->lengths[1]; j--)
            in[j] = 32;
    else
        for (j = INET_ADDRLEN - 1 ; j >= (int)args->lengths[1]; j--)
            in[j] = 32;
    
    family = AF_INET;
    prefix = "";
    if (args->args[0] == NULL) {
        if (args->lengths[1] > INET_ADDRLEN)
            if (memcmp(&in, "\0\0\0\0\0\0\0\0\0\0\0\0", 12) &&
                    memcmp(&in, "\0\0\0\0\0\0\0\0\0\0\377\377", 12)) {
                addr = (char *)&in;
                family = AF_INET6;
            } else
                addr = (char *)&in + 12;
        else
            addr = (char *)&in;
    } else if (*args->args[0] == AF_INET) {
        if (args->lengths[1] > INET_ADDRLEN) {
            if (memcmp(&in, "\0\0\0\0\0\0\0\0\0\0\0\0", 12) && 0 &&
                    memcmp(&in, "\0\0\0\0\0\0\0\0\0\0\377\377", 12)) {
                *null_value = 1;
                return 0;
            } else 
                addr = (char *)&in + 12;
        } else 
            addr = (char *)&in;
    } else if (*args->args[0] == AF_INET6 || *args->args[0] == 10) {
        if (args->lengths[1] <= INET_ADDRLEN) {
            prefix = "::ffff:";
        } else
            family = AF_INET6;

        addr = (char *)&in;
    } else {
        *null_value = 1;
        return 0;
    }

    /* convert */
    if (!inet_ntop(family, addr, out, INET6_ADDRSTRLEN+1)) {
        *null_value = 1;
        return 0;
    }

    result = strcpy(result,prefix);
    result = strcat(result, out);
    *res_length = strlen(result);
    return result;
}

void wfwl_inet_ntop_deinit(UDF_INIT *initid) {

    free(initid->ptr);

}

/**
 * wfwl_inet_contains()
 *
 * Compares IPv4 or IPv6 VARBINARY(16) string with test value, possibly including netmask.
 *
 * @arg1    integer  address family, AF_INET or AF_INET6
 * @arg2    string   varbinary format ipv4 or ipv6 address
 * @arg3    string   human readable ipv4 or ipv6 address, including possibe netmask or prefix bits
 * @return  boolean  true (1) or false (0)
 *
 * Examples: SELECT wfwl_inet_contains(10, ip_saddr_bin, 'fe80::210:eced:1:2337/64')
 *           SELECT wfwl_inet_contains(2, ip_saddr_bin, '1.2.3.4/255.255.0.0')
 *           SELECT wfwl_inet_contains(2, ip_saddr_bin, '1.2.3.4/24')
 */
my_bool wfwl_inet_contains_init(UDF_INIT *initid, UDF_ARGS *args, char *message)
{
    if (args->arg_count != 3 || args->arg_type[0] != INT_RESULT
                             || args->arg_type[1] != STRING_RESULT
                             || args->arg_type[2] != STRING_RESULT) {
        strcpy(message, "Bad args: wfwl_inet_contains(Address family, 4/16 byte binary str, compare string)");
        return 1;
    }
    initid->max_length = INET6_ADDRSTRLEN; /* max length of ipv6 presentation string */
    initid->maybe_null = 0;
    initid->const_item = 1;
    args->maybe_null[0] = 1;
    args->maybe_null[1] = 1;
    args->maybe_null[2] = 0;
    return 0;
}

long long wfwl_inet_contains(UDF_INIT *initid, UDF_ARGS *args, char *result,
        unsigned long *res_length, char *null_value, char *error)
{
    char addr[INET6_ADDRSTRLEN + 4 + 1];
    char buf[INET6_ADDRSTRLEN + 1];
    unsigned int length;
    int family;
    char *ptr;
    unsigned char in4in6 = 0;
    struct in_addr inp;
    struct in6_addr inp6;
    uint32_t netmask = -1;
    uint32_t in_addr_tst = 0;
    uint32_t in_addr_cmp = 0;
    long int bits;
    long int match;

    if (args->args[0] != NULL && !(*args->args[0] == AF_INET || *args->args[0] == AF_INET6 || *args->args[0] == 10))
        return 0;

    if (args->args[1] == NULL || !(args->lengths[1] == INET6_ADDRLEN || args->lengths[1] == INET_ADDRLEN))
        return 0;

    if (!args->args[2] || !(length = args->lengths[2]))
        return 0;

    /* cannot assume null-terminated string according to manual */
    if (length >= sizeof(addr))
        length = sizeof(addr) - 1;
    memcpy(addr, args->args[2], length);
    addr[length] = '\0';

    if (args->args[0] == NULL || *args->args[0] == AF_INET) {
        family = AF_INET;
        bits = 32;
        if (args->lengths[1] == INET6_ADDRLEN)
            in4in6 = 12;
        memcpy(&in_addr_tst, args->args[1] + in4in6, INET_ADDRLEN);
    } else if (*args->args[0] == AF_INET6 || *args->args[0] == 10) {
        family = AF_INET6;
        bits = 128;
        if (args->lengths[1] == INET_ADDRLEN)
            return 0;
    } else
        return 0;

    if (inet_ntop(family, args->args[1] + in4in6, (char *) &buf, INET6_ADDRSTRLEN) == NULL) /* error */
        return 0;

    ptr = strpbrk((char *)&addr, "/");
    if (ptr != NULL) {
        *ptr = '\0';
    }

    opt.verbose = VERBOSE_OFF;
    if (convert_ip(family, addr, &inp, &inp6, NULL)) /* error */
        return 0;

    if (ptr != NULL) {
        if (family == AF_INET) {
            if (strpbrk(++ptr, ".") == NULL) {
                bits = atoi(ptr);
                if (bits < 0 || bits > 32) /* error */
                   return 0;
                netmask = ~(uint32_t)(pow(2, 32 - bits) - 1);
            } else {
                netmask = inet_network(ptr);
            }
            in_addr_cmp = ntohl(inp.s_addr) & netmask;
        } else {
            bits = atoi(++ptr);
            if (bits < 0 || bits > 128) /* error */
               return 0;
       }
    } else {
        if (family == AF_INET)
            in_addr_cmp = ntohl(inp.s_addr);
    }

    if (family == AF_INET) {
        if ((ntohl(in_addr_tst) & netmask) == in_addr_cmp)
            return 1;
    } else if (family == AF_INET6) {
        match = bitncmp(args->args[1], &inp6.s6_addr, bits);
        if (match == 0) /* match */
            return 1;
    }

    return 0;
}

