/* $Id: parser.c 696 2019-09-20 16:47:49Z bhockney $ */

/* (C) 2004-2019 by Bob Hockney <zeus@ix.netcom.com>    *
 * Based on fwlogwatch written by                       *
 * Boris Wesslowski <Wesslowski@CERT.Uni-Stuttgart.DE>  *
 *                                                      *
 * syslog parser for wfwl_syslog                        *
 *                                                      *
 * This code is distributed under the terms of GNU GPL  */

#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <errno.h>
#include <sys/stat.h>
#include <math.h>

#include <main.h>
#include <ipchains.h>
#include <netfilter.h>
#include <ipfilter.h>
#include <win_xp.h>
#include <ipfw.h>
#include <cisco_ios.h>
#include <cisco_pix.h>
#include <snort.h>
#include <netscreen.h>
#include <parser.h>
#include <compare.h>
#include <utils.h>

extern struct options opt;
extern struct selection *selection;
extern struct conn_data *first;
extern struct fields_used fields_used;
unsigned long int rowid = 0;
struct conn_data_sum *data_sum = NULL;


#ifdef HAVE_MAXMINDDB_H
#ifdef HAVE_DB_H
DBT key, data;
#endif
int get_mmdb_string(MMDB_entry_data_s *entry_data, char *ptr);

/* Keep order of these arrays in sync with the enum lists and *
 * structs geoip_locations and geoip asns in main.h           */
static const char * l_paths[GEOIP2_LOC_MAX][5] = {
    {"continent","code",NULL,NULL,NULL},
    {"continent","names",opt.geoip2_locale,NULL,NULL},
    {"country","iso_code",NULL,NULL,NULL},
    {"country","names",opt.geoip2_locale,NULL,NULL},
    {"subdivisions","0","iso_code",NULL,NULL},
    {"subdivisions","0","names",opt.geoip2_locale,NULL},
    {"subdivisions","1","iso_code",NULL,NULL},
    {"subdivisions","1","names",opt.geoip2_locale,NULL},
    {"city","names",opt.geoip2_locale,NULL,NULL},
    {"location","metro_code",NULL,NULL,NULL},
    {"location","time_zone",NULL,NULL,NULL},
    {"registered_country","names",opt.geoip2_locale,NULL,NULL},
    {"represented_country","names",opt.geoip2_locale,NULL,NULL},
    {"traits","is_anonymous_proxy",NULL,NULL,NULL},
    {"traits","is_satellite_provider",NULL,NULL,NULL},
    {"postal","code",NULL,NULL,NULL},
    {"location","latitude",NULL,NULL,NULL},
    {"location","longitude",NULL,NULL,NULL},
    {"country","is_in_european_union",NULL,NULL,NULL},
    {"location","accuracy_radius",NULL,NULL,NULL}
};
static const char * a_paths[GEOIP2_ASN_MAX][5] = {
    {"autonomous_system_number",NULL,NULL,NULL,NULL},
    {"autonomous_system_organization",NULL,NULL,NULL,NULL}
};
#endif

/* adds detail for summarized lines for populating cache */
void add_data_sum(struct log_line *line, unsigned long int rownum) {
  struct conn_data_sum *sum;

  sum = xmalloc(sizeof(struct conn_data_sum));
  sum->rowid = rownum;
  sum->protocol = line->protocol;
  sum->shost = line->shost;
#if HAVE_INET_NTOP
  sum->shost6 = line->shost6;
#endif
  xstrncpy(sum->shost_s, line->shost_s, INET6_ADDRSTRLEN);
  sum->dhost = line->dhost;
#if HAVE_INET_NTOP
  sum->dhost6 = line->dhost6;
#endif
  xstrncpy(sum->dhost_s, line->dhost_s, INET6_ADDRSTRLEN);
  sum->sport = line->sport;
  sum->dport = line->dport;

  sum->next = data_sum;
  data_sum = sum; 
}

/* adds a row to data for output */
void add_entry()
{
  struct conn_data *data;

  data = xmalloc(sizeof(struct conn_data));

  xstrncpy(data->filename, opt.line->filename, FILESIZE);
  data->rowid = ++rowid;
  data->linenum = opt.line->linenum;
  data->count = opt.line->count;
  data->local_time = opt.line->time;
  data->earliest = opt.line->time;
  data->latest = opt.line->time;
  xstrncpy(data->hostname, opt.line->hostname, SHOSTLEN);
  xstrncpy(data->log_label, opt.line->log_label, SHORTLEN);
  xstrncpy(data->branchname, opt.line->branchname, SHORTLEN);
  xstrncpy(data->interface, opt.line->interface, SHORTLEN); /* drop this line */
  data->protocol = opt.line->protocol;
  data->totlen = opt.line->totlen;
  data->family = opt.line->family;
  data->shost = opt.line->shost;
#ifdef HAVE_INET_NTOP
  data->shost6 = opt.line->shost6;
#endif
  xstrncpy(data->shost_s, opt.line->shost_s, INET6_ADDRSTRLEN);
  xstrncpy(data->shostname, opt.line->shostname, SHORTLEN);
  data->sport = opt.line->sport;
  xstrncpy(data->sservice, opt.line->sservice, SHORTLEN);
  data->dhost = opt.line->dhost;
#ifdef HAVE_INET_NTOP
  data->dhost6 = opt.line->dhost6;
#endif
  xstrncpy(data->dhost_s, opt.line->dhost_s, INET6_ADDRSTRLEN);
  xstrncpy(data->dhostname, opt.line->dhostname, SHORTLEN);
  data->dport = opt.line->dport;
  xstrncpy(data->dservice, opt.line->dservice, SHORTLEN);
  data->flags = opt.line->flags;
  data->oob_protocol = opt.line->oob_protocol;
  xstrncpy(data->mac_saddr_str, opt.line->mac_saddr_str, SHORTLEN);
  xstrncpy(data->mac_daddr_str, opt.line->mac_daddr_str, SHORTLEN);
  xstrncpy(data->raw_mac, opt.line->raw_mac, MACSIZE);
  data->fwmark = opt.line->fwmark;
  xstrncpy(data->inif, opt.line->inif, SHORTLEN);
  xstrncpy(data->outif, opt.line->outif, SHORTLEN);
  data->tos = opt.line->tos;
  data->ttl = opt.line->ttl;
  data->ihl = opt.line->ihl;
  data->csum = opt.line->csum;
  data->ipid = opt.line->ipid;
  data->fragoff = opt.line->fragoff;
  data->pay_len = opt.line->pay_len;
  data->flowlabel = opt.line->flowlabel;
  data->tcp_seq = opt.line->tcp_seq;
  data->tcp_ack_seq = opt.line->tcp_ack_seq;
  data->tcp_window = opt.line->tcp_window;
  data->tcp_urgp = opt.line->tcp_urgp;
  data->udp_len = opt.line->udp_len;
  data->icmp_type = opt.line->icmp_type;
  data->icmp_code = opt.line->icmp_code;
  data->icmp_echoid = opt.line->icmp_echoid;
  data->icmp_echoseq = opt.line->icmp_echoseq;
  data->icmp_gw = opt.line->icmp_gw;
  data->icmp_mtu = opt.line->icmp_mtu;
  data->ahesp_spi = opt.line->ahesp_spi;
  data->geoip_locations_src = opt.line->geoip_locations_src;
  data->geoip_locations_dst = opt.line->geoip_locations_dst;
  data->geoip_asns_src = opt.line->geoip_asns_src;
  data->geoip_asns_dst = opt.line->geoip_asns_dst;

  if (opt.resolve_hosts & CACHE_POPULATE_FULL)
    add_data_sum(opt.line, rowid);

  data->next = first;
  first = data;

  opt.report_rows++;
}

/* called when parsed line is not excluded by specification */
void build_list()
{
  struct conn_data *this;
  char stime[TIMESIZE];
  int j;

  /* first (file scope, static) contains list to output */
  /* if new row matches row in struct first, count is incremented by opt.line->count (after repeats) */
  /* if new row does not match, it is added to struct first with count = opt.line->count */
  /* Summarization is done here */

  if (opt.mode == LOG_SUMMARY) {

  this = first;
  while (this != NULL) {
    if ((fields_used.shost) && this->family == AF_INET && (this->shost.s_addr != opt.line->shost.s_addr)) {goto no_match;}
    if ((fields_used.dhost) && this->family == AF_INET && (this->dhost.s_addr != opt.line->dhost.s_addr)) {goto no_match;}
#ifdef HAVE_INET_PTON
    if ((fields_used.shost) && this->family == AF_INET6 && memcmp(this->shost6.s6_addr, opt.line->shost6.s6_addr, sizeof(struct in6_addr))) {goto no_match;}
    if ((fields_used.dhost) && this->family == AF_INET6 && memcmp(this->dhost6.s6_addr, opt.line->dhost6.s6_addr, sizeof(struct in6_addr))) {goto no_match;}
#endif
    if ((fields_used.dport) && (this->dport != opt.line->dport)) {goto no_match;}
    if ((fields_used.sport) && (this->sport != opt.line->sport)) {goto no_match;}
    if ((fields_used.protocol) && (this->protocol != opt.line->protocol)) {goto no_match;}
    if ((fields_used.flags) && (this->flags != opt.line->flags)) {goto no_match;}
    if ((fields_used.inif) && (strncmp (this->interface, opt.line->interface, SHORTLEN) != 0)) {goto no_match;}
    if ((fields_used.log_label) && (strncmp (this->log_label, opt.line->log_label, SHORTLEN) != 0)) {goto no_match;}
    if ((fields_used.hostname) && (strncmp (this->hostname, opt.line->hostname, SHOSTLEN) != 0)) {goto no_match;}
    if ((fields_used.outif) && (strncmp(this->outif, opt.line->outif, SHORTLEN) != 0)) {goto no_match;}
    if ((fields_used.fwmark) && (this->fwmark != opt.line->fwmark)) {goto no_match;}
    if ((fields_used.oob_protocol) && (this->oob_protocol != opt.line->oob_protocol)) {goto no_match;}
    if ((fields_used.mac_saddr_str) && (strncmp(this->mac_saddr_str, opt.line->mac_saddr_str, SHORTLEN) != 0)) {goto no_match;}
    if ((fields_used.mac_daddr_str) && (strncmp(this->mac_daddr_str, opt.line->mac_daddr_str, SHORTLEN) != 0)) {goto no_match;}
    if ((fields_used.raw_mac) && (strncmp(this->raw_mac, opt.line->raw_mac, MACSIZE) != 0)) {goto no_match;}
    if ((fields_used.sname) && (strncmp(this->filename, opt.line->filename, SHORTLEN) != 0)) {goto no_match;}
    if ((fields_used.ipid) && (this->ipid != opt.line->ipid)) {goto no_match;}
    if ((fields_used.csum) && (this->csum != opt.line->csum)) {goto no_match;}
    if ((fields_used.fragoff) && (this->fragoff != opt.line->fragoff)) {goto no_match;}
    if ((fields_used.ihl) && (this->ihl != opt.line->ihl)) {goto no_match;}
    if ((fields_used.tos) && (this->tos != opt.line->tos)) {goto no_match;}
    if ((fields_used.ttl) && (this->ttl != opt.line->ttl)) {goto no_match;}
    if ((fields_used.pay_len) && (this->pay_len != opt.line->pay_len)) {goto no_match;}
    if ((fields_used.flowlabel) && (this->flowlabel != opt.line->flowlabel)) {goto no_match;}
    if ((fields_used.icmp_type) && (this->icmp_type != opt.line->icmp_type)) {goto no_match;}
    if ((fields_used.icmp_code) && (this->icmp_code != opt.line->icmp_code)) {goto no_match;}
    if ((fields_used.totlen) && (this->totlen != opt.line->totlen)) {goto no_match;}
    if ((fields_used.tcp_seq) && (this->tcp_seq != opt.line->tcp_seq)) {goto no_match;}
    if ((fields_used.tcp_ack_seq) && (this->tcp_ack_seq != opt.line->tcp_ack_seq)) {goto no_match;}
    if ((fields_used.tcp_window) && (this->tcp_window != opt.line->tcp_window)) {goto no_match;}
    if ((fields_used.tcp_urgp) && (this->tcp_urgp != opt.line->tcp_urgp)) {goto no_match;}
    if ((fields_used.udp_len) && (this->udp_len != opt.line->udp_len)) {goto no_match;}
    if ((fields_used.icmp_echoid) && (this->icmp_echoid != opt.line->icmp_echoid)) {goto no_match;}
    if ((fields_used.icmp_echoseq) && (this->icmp_echoseq != opt.line->icmp_echoseq)) {goto no_match;}
    if ((fields_used.icmp_gw) && (this->icmp_gw.s_addr != opt.line->icmp_gw.s_addr)) {goto no_match;}
    if ((fields_used.icmp_mtu) && (this->icmp_mtu != opt.line->icmp_mtu)) {goto no_match;}
    if ((fields_used.ahesp_spi) && (this->ahesp_spi != opt.line->ahesp_spi)) {goto no_match;}
    if ((fields_used.continent_code_s) && (strncmp(this->geoip_locations_src->continent_code, opt.line->geoip_locations_src->continent_code, SHORTLEN) != 0)) {goto no_match;}
    if ((fields_used.continent_name_s) && (strncmp(this->geoip_locations_src->continent_name, opt.line->geoip_locations_src->continent_name, SHORTLEN) != 0)) {goto no_match;}
    if ((fields_used.country_iso_code_s) && (strncmp(this->geoip_locations_src->country_iso_code, opt.line->geoip_locations_src->country_iso_code, SHORTLEN) != 0)) {goto no_match;}
    if ((fields_used.country_name_s) && (strncmp(this->geoip_locations_src->country_name, opt.line->geoip_locations_src->country_name, SHORTLEN) != 0)) {goto no_match;}
    if ((fields_used.subdivision_1_iso_code_s) && (strncmp(this->geoip_locations_src->subdivision_1_iso_code, opt.line->geoip_locations_src->subdivision_1_iso_code, SHORTLEN) != 0)) {goto no_match;}
    if ((fields_used.subdivision_1_name_s) && (strncmp(this->geoip_locations_src->subdivision_1_name, opt.line->geoip_locations_src->subdivision_1_name, SHORTLEN) != 0)) {goto no_match;}
    if ((fields_used.subdivision_2_iso_code_s) && (strncmp(this->geoip_locations_src->subdivision_2_iso_code, opt.line->geoip_locations_src->subdivision_2_iso_code, SHORTLEN) != 0)) {goto no_match;}
    if ((fields_used.subdivision_2_name_s) && (strncmp(this->geoip_locations_src->subdivision_2_name, opt.line->geoip_locations_src->subdivision_2_name, SHORTLEN) != 0)) {goto no_match;}
    if ((fields_used.city_name_s) && (strncmp(this->geoip_locations_src->city_name, opt.line->geoip_locations_src->city_name, SHORTLEN) != 0)) {goto no_match;}
    if ((fields_used.metro_code_s) && (this->geoip_locations_src->metro_code != opt.line->geoip_locations_src->metro_code)) {goto no_match;}
    if ((fields_used.accuracy_radius_s) && (this->geoip_locations_src->accuracy_radius != opt.line->geoip_locations_src->accuracy_radius)) {goto no_match;}
    if ((fields_used.time_zone_s) && (strncmp(this->geoip_locations_src->time_zone, opt.line->geoip_locations_src->time_zone, SHORTLEN) != 0)) {goto no_match;}
    if ((fields_used.registered_country_geoname_id_s) && (strncmp(this->geoip_locations_src->registered_country_geoname_id, opt.line->geoip_locations_src->registered_country_geoname_id, SHORTLEN) != 0)) {goto no_match;}
    if ((fields_used.represented_country_geoname_id_s) && (strncmp(this->geoip_locations_src->represented_country_geoname_id, opt.line->geoip_locations_src->represented_country_geoname_id, SHORTLEN) != 0)) {goto no_match;}
    if ((fields_used.is_anonymous_proxy_s) && (this->geoip_locations_src->is_anonymous_proxy != opt.line->geoip_locations_src->is_anonymous_proxy)) {goto no_match;}
    if ((fields_used.is_satellite_provider_s) && (this->geoip_locations_src->is_satellite_provider != opt.line->geoip_locations_src->is_satellite_provider)) {goto no_match;}
    if ((fields_used.is_in_european_union_s) && (this->geoip_locations_src->is_in_european_union != opt.line->geoip_locations_src->is_in_european_union)) {goto no_match;}
    if ((fields_used.postal_code_s) && (strncmp(this->geoip_locations_src->postal_code, opt.line->geoip_locations_src->postal_code, SHORTLEN) != 0)) {goto no_match;}
    if ((fields_used.latitude_s) && (this->geoip_locations_src->latitude != opt.line->geoip_locations_src->latitude)) {goto no_match;}
    if ((fields_used.longitude_s) && (this->geoip_locations_src->longitude != opt.line->geoip_locations_src->longitude)) {goto no_match;}
    if ((fields_used.autonomous_system_number_s) && (this->geoip_asns_src->autonomous_system_number != opt.line->geoip_asns_src->autonomous_system_number)) {goto no_match;}
    if ((fields_used.autonomous_system_organization_s) && (strncmp(this->geoip_asns_src->autonomous_system_organization, opt.line->geoip_asns_src->autonomous_system_organization, SHORTLEN) != 0)) {goto no_match;}
    if ((fields_used.continent_code_d) && (strncmp(this->geoip_locations_dst->continent_code, opt.line->geoip_locations_dst->continent_code, SHORTLEN) != 0)) {goto no_match;}
    if ((fields_used.continent_name_d) && (strncmp(this->geoip_locations_dst->continent_name, opt.line->geoip_locations_dst->continent_name, SHORTLEN) != 0)) {goto no_match;}
    if ((fields_used.country_iso_code_d) && (strncmp(this->geoip_locations_dst->country_iso_code, opt.line->geoip_locations_dst->country_iso_code, SHORTLEN) != 0)) {goto no_match;}
    if ((fields_used.country_name_d) && (strncmp(this->geoip_locations_dst->country_name, opt.line->geoip_locations_dst->country_name, SHORTLEN) != 0)) {goto no_match;}
    if ((fields_used.subdivision_1_iso_code_d) && (strncmp(this->geoip_locations_dst->subdivision_1_iso_code, opt.line->geoip_locations_dst->subdivision_1_iso_code, SHORTLEN) != 0)) {goto no_match;}
    if ((fields_used.subdivision_1_name_d) && (strncmp(this->geoip_locations_dst->subdivision_1_name, opt.line->geoip_locations_dst->subdivision_1_name, SHORTLEN) != 0)) {goto no_match;}
    if ((fields_used.subdivision_2_iso_code_d) && (strncmp(this->geoip_locations_dst->subdivision_2_iso_code, opt.line->geoip_locations_dst->subdivision_2_iso_code, SHORTLEN) != 0)) {goto no_match;}
    if ((fields_used.subdivision_2_name_d) && (strncmp(this->geoip_locations_dst->subdivision_2_name, opt.line->geoip_locations_dst->subdivision_2_name, SHORTLEN) != 0)) {goto no_match;}
    if ((fields_used.city_name_d) && (strncmp(this->geoip_locations_dst->city_name, opt.line->geoip_locations_dst->city_name, SHORTLEN) != 0)) {goto no_match;}
    if ((fields_used.metro_code_d) && (this->geoip_locations_dst->metro_code != opt.line->geoip_locations_dst->metro_code)) {goto no_match;}
    if ((fields_used.accuracy_radius_d) && (this->geoip_locations_dst->accuracy_radius != opt.line->geoip_locations_dst->accuracy_radius)) {goto no_match;}
    if ((fields_used.time_zone_d) && (strncmp(this->geoip_locations_dst->time_zone, opt.line->geoip_locations_dst->time_zone, SHORTLEN) != 0)) {goto no_match;}
    if ((fields_used.registered_country_geoname_id_d) && (strncmp(this->geoip_locations_dst->registered_country_geoname_id, opt.line->geoip_locations_dst->registered_country_geoname_id, SHORTLEN) != 0)) {goto no_match;}
    if ((fields_used.represented_country_geoname_id_d) && (strncmp(this->geoip_locations_dst->represented_country_geoname_id, opt.line->geoip_locations_dst->represented_country_geoname_id, SHORTLEN) != 0)) {goto no_match;}
    if ((fields_used.is_anonymous_proxy_d) && (this->geoip_locations_dst->is_anonymous_proxy != opt.line->geoip_locations_dst->is_anonymous_proxy)) {goto no_match;}
    if ((fields_used.is_satellite_provider_d) && (this->geoip_locations_dst->is_satellite_provider != opt.line->geoip_locations_dst->is_satellite_provider)) {goto no_match;}
    if ((fields_used.is_in_european_union_d) && (this->geoip_locations_dst->is_in_european_union != opt.line->geoip_locations_dst->is_in_european_union)) {goto no_match;}
    if ((fields_used.postal_code_d) && (strncmp(this->geoip_locations_dst->postal_code, opt.line->geoip_locations_dst->postal_code, SHORTLEN) != 0)) {goto no_match;}
    if ((fields_used.latitude_d) && (this->geoip_locations_dst->latitude != opt.line->geoip_locations_dst->latitude)) {goto no_match;}
    if ((fields_used.longitude_d) && (this->geoip_locations_dst->longitude != opt.line->geoip_locations_dst->longitude)) {goto no_match;}
    if ((fields_used.autonomous_system_number_d) && (this->geoip_asns_dst->autonomous_system_number != opt.line->geoip_asns_dst->autonomous_system_number)) {goto no_match;}
    if ((fields_used.autonomous_system_organization_d) && (strncmp(this->geoip_asns_dst->autonomous_system_organization, opt.line->geoip_asns_dst->autonomous_system_organization, SHORTLEN) != 0)) {goto no_match;}


    if (opt.line->time >= this->latest || opt.line->time <= this->earliest) {
      if (opt.line->time >= this->latest) {
        this->latest = opt.line->time;
      }
      if (opt.line->time <= this->earliest) {
        this->earliest = opt.line->time;
      }
    } else {
      if(opt.verbose >= VERBOSE_WARNING
        && strcmp(this->filename, opt.line->filename) == 0) {
        strftime(stime, TIMESIZE, "%b %d %H:%M:%S", localtime(&this->latest));
        fprintf(stderr, "Timewarp in log file (%s", stime);
        strftime(stime, TIMESIZE, "%b %d %H:%M:%S", localtime(&opt.line->time));
        fprintf(stderr, " > %s).\n", stime);
      }
    }

    this->count += opt.line->count;

    for (j = opt.line->count; j > 1; j--) {
      opt.matched_entries++;
    }

    if (opt.resolve_hosts & CACHE_POPULATE_FULL)
      add_data_sum(opt.line, this->rowid);
    return;

    no_match:
    this = this->next;
  } /* end while (this != NULL) */

  }

  for (j = opt.line->count; j > 1; j--) {
    opt.matched_entries++;
  }

  if (opt.mode == LOG_SUMMARY) {
    add_entry();
  } else {
    for (j = opt.line->count; j > 0; j--) {
      opt.line->count = 1;
      add_entry();
    }
  }
}

#ifdef HAVE_MAXMINDDB_H
int get_mmdb_string(MMDB_entry_data_s *entry_data, char *ptr) {
  unsigned int data_size = SHORTLEN;

  *ptr = '\0';

  if(!entry_data->has_data)
    return 0;

  if (entry_data->data_size < data_size)
    data_size=entry_data->data_size +1;

  xstrncpy(ptr,entry_data->utf8_string,data_size);

  return 0;
}
#endif

/* populates geoip data for source ip and/or dest ip as needed.     *
 * use bdb to track pointers after geoip lookup for each unique IP, *
 * so only lookup once for each IP.                                 */
int geoip_populate(unsigned char type) {
#ifdef HAVE_MAXMINDDB_H
  char ipstr[INET6_ADDRSTRLEN];
  int gai_error, mmdb_error, status, entry_key;
  struct geoip_locations *ldata, **ldataptr = NULL;
  struct geoip_asns *adata, **adataptr = NULL;
  unsigned char poploc;
  MMDB_s * geoip2_handle;
  MMDB_lookup_result_s result;
  MMDB_entry_data_list_s *entry_data_list = NULL;
  MMDB_entry_data_s entry_data;
#ifdef HAVE_DB_H
  DB *dbp = NULL;
  unsigned char prefix;
  uint32_t k;
  struct in6_addr in_addr_ptr;
  int ret;
  char net_string[INET6_ADDRSTRLEN +4 ];
  char net_prefix[5];
#endif
#endif
  switch (type) {
#ifdef HAVE_MAXMINDDB_H
    case GEOIP_LOC_SRC:
      xstrncpy(ipstr, opt.line->shost_s, INET6_ADDRSTRLEN);
      ldataptr = &opt.line->geoip_locations_src;
#ifdef HAVE_DB_H
      dbp = opt.dbp_locations;
#endif
      geoip2_handle = opt.geoip2_locations_handle;
      poploc = true;
      break;
    case GEOIP_LOC_DST:
      xstrncpy(ipstr, opt.line->dhost_s, INET6_ADDRSTRLEN);
      ldataptr = &opt.line->geoip_locations_dst;
#ifdef HAVE_DB_H
      dbp = opt.dbp_locations;
#endif
      geoip2_handle = opt.geoip2_locations_handle;
      poploc = true;
      break;
    case GEOIP_ASN_SRC:
      xstrncpy(ipstr, opt.line->shost_s, INET6_ADDRSTRLEN);
      adataptr = &opt.line->geoip_asns_src;
#ifdef HAVE_DB_H
      dbp = opt.dbp_asns;
#endif
      geoip2_handle = opt.geoip2_asns_handle;
      poploc = false;
      break;
    case GEOIP_ASN_DST:
      xstrncpy(ipstr, opt.line->dhost_s, INET6_ADDRSTRLEN);
      adataptr = &opt.line->geoip_asns_dst;
#ifdef HAVE_DB_H
      dbp = opt.dbp_asns;
#endif
      geoip2_handle = opt.geoip2_asns_handle;
      poploc = false;
      break;
#endif
    default:
      return 0;
      break;
  }

#ifdef HAVE_MAXMINDDB_H
  if (geoip2_handle != NULL) {

  result = MMDB_lookup_string(geoip2_handle, ipstr, &gai_error, &mmdb_error);

  if (gai_error != 0) {
    if(opt.verbose >= VERBOSE_NOTICE)
      fprintf(stderr, "Error from getaddrinfo for %s: %s\n",
          (char *)&ipstr, gai_strerror(gai_error));

    return 0;
  }

  if (mmdb_error != MMDB_SUCCESS) {
    if(opt.verbose >= VERBOSE_INFO)
      fprintf(stderr,"libmaxminddb error: %s\n", MMDB_strerror(mmdb_error));

    return 0;
  }

  if (result.found_entry) {
    status = MMDB_get_entry_data_list(&result.entry, &entry_data_list);
    if (status != MMDB_SUCCESS) {
      if(opt.verbose >= VERBOSE_INFO)
        fprintf(stderr, "Error looking up maxminddb entry data: %s\n", MMDB_strerror(status));

      goto cleanup;
    }
  } else {
    if (opt.verbose >= VERBOSE_DEBUG)
      fprintf(stderr, "No maxminddb entry for IP address: %s\n", (char *)&ipstr);

    goto cleanup;
  }
#ifdef HAVE_DB_H
  ret = convert_ip(opt.line->family, (char *)&ipstr, (struct in_addr *)&in_addr_ptr, &in_addr_ptr, NULL);

  if (opt.line->family == AF_INET) {
    prefix = result.netmask % 32;
    k = pow(2,32) - 1;
    k <<= 32-prefix;
    *(uint32_t *)&in_addr_ptr = *(uint32_t *)&in_addr_ptr & htonl(k);
  } else if(opt.line->family == AF_INET6) {
    prefix = result.netmask;
    zero_after_n_bits (&in_addr_ptr, prefix, sizeof(struct in6_addr));
  } else {
    goto cleanup;
  }

  snprintf((char *)&net_prefix, sizeof(net_prefix), "/%-3d", prefix);
  if(inet_ntop(opt.line->family, &in_addr_ptr, (char *)&net_string, INET6_ADDRSTRLEN) != NULL) {
    strncat((char *)&net_string, (char *)&net_prefix, INET6_ADDRSTRLEN + 4 - strlen((char *)&net_string) - 1);
  if(opt.verbose >= VERBOSE_DEBUG)
    fprintf(stderr, "cidr: %s\n", (char *)&net_string);
  }

  key.data = &net_string;
  key.size = strlen((char *)&net_string) + 1;

  ret = dbp->get(dbp, NULL, &key, &data, 0);
  switch (ret) {
    case 0:
      /* key found */
      if (opt.verbose >= VERBOSE_DEBUG)
        fprintf(stderr, "Key found: %s, data %p\n", (char *)key.data, *(void **)data.data);
      if (poploc) {
        *ldataptr = *(struct geoip_locations **)data.data;
      } else {
        *adataptr = *(struct geoip_asns **)data.data;
      }
      break;
    case DB_NOTFOUND:
#endif
      if (poploc) { /* getting locations data */

      ldata = xmalloc(sizeof(struct geoip_locations));
      memcpy(ldata, opt.geoip_locations_init, sizeof(struct geoip_locations));
 
      for (entry_key=0; entry_key<GEOIP2_LOC_MAX; entry_key++) {
        MMDB_aget_value(&result.entry, &entry_data, l_paths[entry_key]);
        switch(entry_key) {
          case GEOIP2_CONT_CODE:
            get_mmdb_string(&entry_data, ldata->continent_code);
            break;
          case GEOIP2_CONT_NAME:
            get_mmdb_string(&entry_data, ldata->continent_name);
            break;
          case GEOIP2_COUNTRY_ISO:
            get_mmdb_string(&entry_data, ldata->country_iso_code);
            break;
          case GEOIP2_COUNTRY_NAME:
            get_mmdb_string(&entry_data, ldata->country_name);
            break;
          case GEOIP2_SUB_0_ISO:
            get_mmdb_string(&entry_data, ldata->subdivision_1_iso_code);
            break;
          case GEOIP2_SUB_0_NAME:
            get_mmdb_string(&entry_data, ldata->subdivision_1_name);
            break;
          case GEOIP2_SUB_1_ISO:
            get_mmdb_string(&entry_data, ldata->subdivision_2_iso_code);
            break;
          case GEOIP2_SUB_1_NAME:
            get_mmdb_string(&entry_data, ldata->subdivision_2_name);
            break;
          case GEOIP2_CITY:
            get_mmdb_string(&entry_data, ldata->city_name);
            break;
          case GEOIP2_METRO:
            if(entry_data.has_data)
              ldata->metro_code = entry_data.uint16;
            break;
          case GEOIP2_TZ:
            get_mmdb_string(&entry_data, ldata->time_zone);
            break;
          case GEOIP2_ACCURACY_RADIUS:
            if(entry_data.has_data)
              ldata->accuracy_radius = entry_data.uint16;
            break;
          case GEOIP2_REG_COUNTRY_NAME:
            get_mmdb_string(&entry_data, ldata->registered_country_geoname_id);
            break;
          case GEOIP2_REP_COUNTRY_NAME:
            get_mmdb_string(&entry_data, ldata->represented_country_geoname_id);
            break;
          case GEOIP2_ANON_PROXY:
            if(entry_data.has_data)
              ldata->is_anonymous_proxy = entry_data.boolean;
            else
              ldata->is_anonymous_proxy = false;
            break;
          case GEOIP2_SAT_PROV:
            if(entry_data.has_data)
              ldata->is_satellite_provider = entry_data.boolean;
            else
              ldata->is_satellite_provider = false;
            break;
          case GEOIP2_IN_EU:
            if(entry_data.has_data)
              ldata->is_in_european_union = entry_data.boolean;
            else
              ldata->is_in_european_union = false;
            break;
          case GEOIP2_POSTAL:
            get_mmdb_string(&entry_data, ldata->postal_code);
            break;
          case GEOIP2_LATITUDE:
            if(entry_data.has_data)
              ldata->latitude = entry_data.double_value;
            break;
          case GEOIP2_LONGITUDE:
            if(entry_data.has_data)
              ldata->longitude = entry_data.double_value;
            break;
          default:
            if (opt.verbose >= VERBOSE_NOTICE)
              fprintf(stderr,"Unrecognized geoip locations entry key: %d\n", entry_key);
            break;
        } /* end switch(entry_key) */
      } /* end for (entry_key=0; entry_key<GEOIP2_LOC_MAX; entry_key++) */
      } else { /* getting asn data */

      adata = xmalloc(sizeof(struct geoip_asns));
      memcpy(adata, opt.geoip_asns_init, sizeof(struct geoip_asns));
 
      for (entry_key=0; entry_key<GEOIP2_ASN_MAX; entry_key++) {
        MMDB_aget_value(&result.entry, &entry_data, a_paths[entry_key]);
        switch(entry_key) {
          case GEOIP2_ASN:
            if(entry_data.has_data)
              adata->autonomous_system_number = entry_data.uint32;
            break;
          case GEOIP2_AS_ORG:
            get_mmdb_string(&entry_data, adata->autonomous_system_organization);
            break;
          default:
            if (opt.verbose >= VERBOSE_NOTICE)
              fprintf(stderr,"Unrecognized geoip asn entry key: %d\n", entry_key);
            break;
        } /* end switch(entry_key) */
      } /* end for (entry_key=0; entry_key<GEOIP2_ASN_MAX; entry_key++) */
      }   
      if (poploc) { 
        *ldataptr = ldata;
      } else {
        *adataptr = adata;
      }
#ifdef HAVE_DB_H
      if (poploc) {
        data.data = &ldata;
        data.size = sizeof(ldata);
      } else {
        data.data = &adata;
        data.size = sizeof(adata);
      }
      ret = dbp->put(dbp, NULL, &key, &data, 0);
      if (opt.verbose >= VERBOSE_DEBUG)
        fprintf(stderr, "Added new key %s, data %p\n", (char *)key.data, *(void **)data.data);
      break;
    default:
      /* error */
      fprintf(stderr, "db_address lookup: %s\n", db_strerror(ret));
      break;
  } /* end switch (ret) */
#endif

  cleanup:

  if (entry_data_list != NULL)
    MMDB_free_entry_data_list(entry_data_list);

  } /* if (geoip2_handle != NULL) */
#endif
  return 0;
}

/* parses a logfile line */
/* called with input = log line and linenum = line number */
unsigned char parse_line(char *input, unsigned long linenum)
{
  unsigned char retval;
  char *pnt, *endpnt;
  struct selection *sel;
#if HAVE_INET_NTOP
  unsigned long int match_inet6;
#endif
  unsigned int match = P_MATCH_NONE;
  uint64_t match_inc = MATCH_MUL_NONE;
  uint64_t match_exc = MATCH_MUL_NONE;
  int res;
  char month[4], ptime[9], name[SHOSTLEN], rest[BUFSIZE];
  int day;
  int repeated = 0;
  unsigned long line_value;
  double double_value;
  char line_boolean;
  char * line_string = NULL;
  unsigned long long match_value = 0;
  bool extra_test;
  struct in_addr host;
#if HAVE_INET_NTOP 
  struct in6_addr host6;
#endif

  /* if repeated message, update count */
  pnt = strstr(input, " last message repeated ");
  if (pnt != NULL) {
    if (opt.repeated == 1) {
      retval = sscanf(input, "%3s %2d %8s %" SHOSTLEN_S "s %" BUFSIZE_S "s", month, &day, ptime, name, rest);
      if (retval == 5) {
        if(strncmp(opt.line->hostname, name, SHOSTLEN) == 0) {
          repeated = atoi(pnt+23);
          /* opt.orig_count is 0 or 1 */
          opt.line->count = opt.orig_count * repeated;
          build_list();
          if (opt.verbose >= VERBOSE_ERROR)
            fprintf(stderr, "r");
          return PARSE_OK;
        }
      }
      if (opt.verbose >= VERBOSE_ERROR)
        fprintf(stderr, "_");
      return PARSE_NO_HIT;
    }
  } else {
    /* rsyslog style repeats */
    repeated = 0;
    pnt = strstr(input, " message repeated ");
    if (pnt != NULL) {
      repeated = atoi(pnt+18);
      endpnt = strchr(pnt, '[');
      if (endpnt != NULL) {
        while (endpnt >= pnt) {
          *endpnt = ' ';
          endpnt--;
        }
      }
      pnt = strrchr(pnt, ']');
      if (pnt != NULL)
        *pnt = ' ';
    }
  }

  opt.linenum = linenum;

  /* try parsers until one works */
  if ((opt.format & PARSER_IPCHAINS) && (strstr(input, " kernel: Packet log: "))) {
    /* For ipchains log format see (in kernel 2.2 source) */
    /* /usr/src/linux/net/ipv4/ip_fw.c */
    retval = flex_ipchains(input, linenum);
  } else if ((opt.format & PARSER_NETFILTER) && (strstr(input, " OUT="))) {
    /* For netfilter log format see (in kernel 2.4 source) */
    /* /usr/src/linux/net/ipv4/netfilter/ipt_LOG.c */
    retval = flex_netfilter(input, linenum);
  } else if ((opt.format & PARSER_CISCO_IOS) && (strstr(input, "%SEC-6-IPACCESSLOG"))) {
    /* For cisco log format see CCO */
    retval = flex_cisco_ios(input, linenum);
  } else if ((opt.format & PARSER_IPFILTER) && (strstr(input, " ipmon["))) {
    /* For ipfilter log format see the source */
    retval = flex_ipfilter(input, linenum);
  } else if ((opt.format & PARSER_IPFW) && (strstr(input, " ipfw: "))) {
    retval = flex_ipfw(input, linenum);
  } else if ((opt.format & PARSER_CISCO_PIX) && ((strstr(input, "%PIX-") || strstr(input, "%FWSM-")))) {
    /* For cisco log format see CCO */
    retval = flex_cisco_pix(input, linenum);
  } else if ((opt.format & PARSER_NETSCREEN) && (strstr(input, " NetScreen "))) {
    retval = flex_netscreen(input, linenum);
  } else if (opt.format & PARSER_WIN_XP){
    retval = win_xp(input, linenum);
  } else if ((opt.format & PARSER_SNORT) && (strstr(input, " snort"))) {
    retval = flex_snort(input, linenum);
  } else {
    retval = PARSE_NO_HIT;
  }

  if (retval == PARSE_NO_HIT) {
    if (opt.verbose >= VERBOSE_ERROR)
      fprintf(stderr, "_");
    return PARSE_NO_HIT;
  }

  /* true if successfully matched by a parser */
  if (retval == PARSE_OK) {

    /* rsyslog style repeats */
    if (repeated)
      opt.line->count = repeated;

    if (!(opt.use_geoip & GEOIP_LOC_OFF) && (opt.use_geoip & GEOIP_LOC_SRC))
      geoip_populate (GEOIP_LOC_SRC);

    if (!(opt.use_geoip & GEOIP_LOC_OFF) && (opt.use_geoip & GEOIP_LOC_DST))
      geoip_populate (GEOIP_LOC_DST);

    if (!(opt.use_geoip & GEOIP_ASN_OFF) && (opt.use_geoip & GEOIP_ASN_SRC))
      geoip_populate (GEOIP_ASN_SRC);

    if (!(opt.use_geoip & GEOIP_ASN_OFF) && (opt.use_geoip & GEOIP_ASN_DST))
      geoip_populate (GEOIP_ASN_DST);

    /* Set match = P_MATCH_EXC to exclude parsed line.            *
     * For fields that allow multiple criteria, all rules must be *
     * tested before verdict can be reached, so match is not used *
     * in favor of match_exc, which is checked after all rules    *
     * have been processed.                                       */

    sel = selection;

    while (sel != NULL && match != P_MATCH_EXC) {
      if (sel->have_value == -1) {
        sel = sel->next;
        continue;
      }
      if(
        ((opt.use_geoip & GEOIP_LOC_OFF) && (sel->geoip & (GEOIP_LOC_SRC | GEOIP_LOC_DST))) ||
        ((opt.use_geoip & GEOIP_ASN_OFF) && (sel->geoip & (GEOIP_ASN_SRC | GEOIP_ASN_DST)))
        ) {
        sel = sel->next;
	continue;
      }

      extra_test = false;

      switch (sel->field) {
        /* Begin numberic types */
        case SEL_OOB_FAMILY:
        case SEL_OOB_PROTOCOL:
        case SEL_PROTOCOL:
        case SEL_IP_TOTLEN:
        case SEL_IP_TOS:
        case SEL_IP_TTL:
        case SEL_IP_IHL:
        case SEL_TCP_SPORT:
        case SEL_TCP_DPORT:
        case SEL_UDP_SPORT:
        case SEL_UDP_DPORT:
        case SEL_ICMP_TYPE:
        case SEL_ICMP_CODE:
        case SEL_ICMP_ECHOID:
        case SEL_ICMP_ECHOSEQ:
        case SEL_ICMP_FRAGMTU:
        case SEL_ICMPv6_TYPE:
        case SEL_ICMPv6_CODE:
        case SEL_ICMPv6_ECHOID:
        case SEL_ICMPv6_ECHOSEQ:
        case SEL_AHESP_SPI:
        case SEL_IP_ID:
        case SEL_IP_CSUM:
        case SEL_IP_FRAGOFF:
        case SEL_IP6_PAY_LEN:
        case SEL_IP6_FLOWLABEL:
        case SEL_OOB_MARK:
        case SEL_TCP_SEQ:
        case SEL_TCP_ACKSEQ:
        case SEL_TCP_WINDOW:
        case SEL_TCP_URGP:
        case SEL_UDP_LEN: 
        case SEL_METRO_CODE_SRC:
        case SEL_ACCURACY_RADIUS_SRC:
        case SEL_METRO_CODE_DST:
        case SEL_ACCURACY_RADIUS_DST:
        case SEL_ASN_SRC:
        case SEL_ASN_DST:
          switch (sel->field) {
            case SEL_OOB_FAMILY:
              line_value = opt.line->family;
              match_value = MATCH_OOB_FAMILY;
              break;
            case SEL_PROTOCOL:
              line_value = opt.line->protocol;
              match_value = MATCH_PROTOCOL;
              break;
            case SEL_IP_TOS:
              line_value = opt.line->tos;
              match_value = MATCH_IP_TOS;
              break;
            case SEL_IP_TTL:
              line_value = opt.line->ttl;
              match_value = MATCH_IP_TTL;
              break;
            case SEL_IP_IHL:
              line_value = opt.line->ihl;
              match_value = MATCH_IP_IHL;
              break;
            case SEL_ICMP_TYPE:
              line_value = opt.line->icmp_type;
              match_value = MATCH_ICMP_TYPE;
              extra_test = opt.line->protocol != 1;
              break;
            case SEL_ICMP_CODE:
              line_value = opt.line->icmp_code;
              match_value = MATCH_ICMP_CODE;
              extra_test = opt.line->protocol != 1;
              break;
            case SEL_ICMPv6_TYPE:
              line_value = opt.line->icmp_type;
              match_value = MATCH_ICMPv6_TYPE;
              extra_test = opt.line->protocol !=58;
              break;
            case SEL_ICMPv6_CODE:
              line_value = opt.line->icmp_code;
              match_value = MATCH_ICMPv6_CODE;
              extra_test = opt.line->protocol !=58;
              break;
            case SEL_OOB_PROTOCOL:
              line_value = opt.line->oob_protocol;
              match_value = MATCH_OOB_PROTOCOL;
              break;
            case SEL_IP_ID:
              line_value = opt.line->ipid;
              match_value = MATCH_IP_ID;
              break;
            case SEL_IP_CSUM:
              line_value = opt.line->csum;
              match_value = MATCH_IP_CSUM;
              break;
            case SEL_IP_FRAGOFF:
              line_value = opt.line->fragoff;
              match_value = MATCH_IP_FRAGOFF;
              break;
            case SEL_IP_TOTLEN:
              line_value = opt.line->totlen;
              match_value = MATCH_IP_TOTLEN;
              break;
            case SEL_IP6_PAY_LEN:
              line_value = opt.line->pay_len;
              match_value = MATCH_IP6_PAY_LEN;
              break;
            case SEL_TCP_WINDOW:
              line_value = opt.line->tcp_window;
              match_value = MATCH_TCP_WINDOW;
              break;
            case SEL_TCP_URGP:
              line_value = opt.line->tcp_urgp;
              match_value = MATCH_TCP_URGP;
              break;
            case SEL_UDP_LEN: 
              line_value = opt.line->udp_len;
              match_value = MATCH_UDP_LEN;
              break;
            case SEL_ICMP_ECHOID:
              line_value = opt.line->icmp_echoid;
              match_value = MATCH_ICMP_ECHOID;
              extra_test = opt.line->protocol != 1 || (opt.line->icmp_type != 8 && opt.line->icmp_type != 0);
              break;
            case SEL_ICMP_ECHOSEQ:
              line_value = opt.line->icmp_echoseq;
              match_value = MATCH_ICMP_ECHOSEQ;
              extra_test = opt.line->protocol != 1 || (opt.line->icmp_type != 8 && opt.line->icmp_type != 0);
              break;
            case SEL_ICMP_FRAGMTU:
              line_value = opt.line->icmp_mtu;
              match_value = MATCH_ICMP_FRAGMTU;
              extra_test = opt.line->protocol != 1 || !(opt.line->icmp_type == 3 && opt.line->icmp_code == 4);
              break;
            case SEL_ICMPv6_ECHOID:
              line_value = opt.line->icmp_echoid;
              match_value = MATCH_ICMPv6_ECHOID;
              extra_test = opt.line->protocol != 58 || (opt.line->icmp_type != 128 && opt.line->icmp_type != 129);
              break;
            case SEL_ICMPv6_ECHOSEQ:
              line_value = opt.line->icmp_echoseq;
              match_value = MATCH_ICMPv6_ECHOSEQ;
              extra_test = opt.line->protocol != 58 || (opt.line->icmp_type != 128 && opt.line->icmp_type != 129);
              break;
            case SEL_OOB_MARK:
              line_value = opt.line->fwmark;
              match_value = MATCH_OOB_MARK;
              break;
            case SEL_IP6_FLOWLABEL:
              line_value = opt.line->flowlabel;
              match_value = MATCH_IP6_FLOWLABEL;
              break;
            case SEL_TCP_SPORT:
              line_value = opt.line->sport;
              match_value = MATCH_TCP_SPORT;
              extra_test = opt.line->protocol !=6;
              break;
            case SEL_TCP_DPORT:
              line_value = opt.line->dport;
              match_value = MATCH_TCP_DPORT;
              extra_test = opt.line->protocol !=6;
              break;
            case SEL_TCP_SEQ:
              line_value = opt.line->tcp_seq;
              match_value = MATCH_TCP_SEQ;
              break;
            case SEL_TCP_ACKSEQ:
              line_value = opt.line->tcp_ack_seq;
              match_value = MATCH_TCP_ACKSEQ;
              break;
            case SEL_UDP_SPORT:
              line_value = opt.line->sport;
              match_value = MATCH_UDP_SPORT;
              extra_test = opt.line->protocol !=17;
              break;
            case SEL_UDP_DPORT:
              line_value = opt.line->dport;
              match_value = MATCH_UDP_DPORT;
              extra_test = opt.line->protocol !=17;
              break;
            case SEL_AHESP_SPI:
              line_value = opt.line->ahesp_spi;
              match_value = MATCH_AHESP_SPI;
              extra_test = opt.line->protocol != 50 && opt.line->protocol != 51;
              break;
            case SEL_METRO_CODE_SRC:
              line_value = opt.line->geoip_locations_src->metro_code;
              match_value = MATCH_METRO_CODE_SRC;
              break;
            case SEL_ACCURACY_RADIUS_SRC:
              line_value = opt.line->geoip_locations_src->accuracy_radius;
              match_value = MATCH_ACCURACY_RADIUS_SRC;
              break;
            case SEL_METRO_CODE_DST:
              line_value = opt.line->geoip_locations_dst->metro_code;
              match_value = MATCH_METRO_CODE_DST;
              break;
            case SEL_ACCURACY_RADIUS_DST:
              line_value = opt.line->geoip_locations_dst->accuracy_radius;
              match_value = MATCH_ACCURACY_RADIUS_DST;
              break;
            case SEL_ASN_SRC:
              line_value = opt.line->geoip_asns_src->autonomous_system_number;
              match_value = MATCH_ASN_SRC;
              break;
            case SEL_ASN_DST:
              line_value = opt.line->geoip_asns_dst->autonomous_system_number;
              match_value = MATCH_ASN_DST;
              break;
            default:
              line_value = 0;
              break;
          }
          if (extra_test || (sel->invert && (match_exc & match_value)) || (!sel->invert && (match_inc & match_value)))
            break;
          if (((line_value < sel->value || line_value > sel->max_value) && !(sel->invert)) \
            || (line_value >= sel->value && line_value <= sel->max_value && (sel->invert)) ) {
            match_exc |= match_value;
            match_inc &= ~match_value;
          } else {
            match_inc |= match_value;
            match_exc &= ~match_value;
          }
          break;
        /* End numeric types */
        /* Begin boolean types */
        case SEL_IS_ANONYMOUS_PROXY_SRC:
        case SEL_IS_SATELLITE_PROVIDER_SRC:
        case SEL_IS_IN_EUROPEAN_UNION_SRC:
        case SEL_IS_ANONYMOUS_PROXY_DST:
        case SEL_IS_SATELLITE_PROVIDER_DST:
        case SEL_IS_IN_EUROPEAN_UNION_DST:
          switch(sel->field) {
            case SEL_IS_ANONYMOUS_PROXY_SRC:
              line_boolean = opt.line->geoip_locations_src->is_anonymous_proxy;
              break;
            case SEL_IS_SATELLITE_PROVIDER_SRC:
              line_boolean = opt.line->geoip_locations_src->is_satellite_provider;
              break;
            case SEL_IS_IN_EUROPEAN_UNION_SRC:
              line_boolean = opt.line->geoip_locations_src->is_in_european_union;
              break;
            case SEL_IS_ANONYMOUS_PROXY_DST:
              line_boolean = opt.line->geoip_locations_dst->is_anonymous_proxy;
              break;
            case SEL_IS_SATELLITE_PROVIDER_DST:
              line_boolean = opt.line->geoip_locations_dst->is_satellite_provider;
              break;
            case SEL_IS_IN_EUROPEAN_UNION_DST:
              line_boolean = opt.line->geoip_locations_dst->is_in_european_union;
              break;
            default:
              line_boolean = false;
              break;
          }
          if (sel->boolean != line_boolean)
            match = P_MATCH_EXC;
          break;
        /* End boolean types */
        /* Begin double types */
        case SEL_LATITUDE_SRC:
        case SEL_LONGITUDE_SRC:
        case SEL_LATITUDE_DST:
        case SEL_LONGITUDE_DST:
          switch(sel->field) {
            case SEL_LATITUDE_SRC:
              double_value = opt.line->geoip_locations_src->latitude;
              match_value = MATCH_LATITUDE_SRC;
              break;
            case SEL_LONGITUDE_SRC:
              double_value = opt.line->geoip_locations_src->longitude;
              match_value = MATCH_LONGITUDE_SRC;
              break;
            case SEL_LATITUDE_DST:
              double_value = opt.line->geoip_locations_dst->latitude;
              match_value = MATCH_LATITUDE_DST;
              break;
            case SEL_LONGITUDE_DST:
              double_value = opt.line->geoip_locations_dst->longitude;
              match_value = MATCH_LONGITUDE_DST;
              break;
            default:
              double_value = 0;
              break;
          }
          if ((sel->invert && (match_exc & match_value)) || (!sel->invert && (match_inc & match_value)))
            break;
          if (((double_value < sel->double_value || double_value > sel->max_double_value) && !(sel->invert)) \
            || (double_value >= sel->double_value && double_value <= sel->max_double_value && (sel->invert)) ) {
            match_exc |= match_value;
            match_inc &= ~match_value;
          } else {
            match_inc |= match_value;
            match_exc &= ~match_value;
          }
          break;
        /* End double types */
        /* Begin Regex types */
        case SEL_PREFIX:
        case SEL_LOCAL_HOST:
        case SEL_IN_IF:
        case SEL_OUT_IF:
        case SEL_MAC_SADDR:
        case SEL_MAC_DADDR:
        case SEL_MAC:
        case SEL_CONTINENT_CODE_SRC:
        case SEL_CONTINENT_NAME_SRC:
        case SEL_COUNTRY_ISO_CODE_SRC:
        case SEL_COUNTRY_NAME_SRC:
        case SEL_SUBDIVISION_1_ISO_CODE_SRC:
        case SEL_SUBDIVISION_1_NAME_SRC:
        case SEL_SUBDIVISION_2_ISO_CODE_SRC:
        case SEL_SUBDIVISION_2_NAME_SRC:
        case SEL_CITY_NAME_SRC:
        case SEL_TIME_ZONE_SRC:
        case SEL_REGISTERED_COUNTRY_GEONAME_ID_SRC:
        case SEL_REPRESENTED_COUNTRY_GEONAME_ID_SRC:
        case SEL_POSTAL_CODE_SRC:
        case SEL_AS_ORG_SRC:
        case SEL_CONTINENT_CODE_DST:
        case SEL_CONTINENT_NAME_DST:
        case SEL_COUNTRY_ISO_CODE_DST:
        case SEL_COUNTRY_NAME_DST:
        case SEL_SUBDIVISION_1_ISO_CODE_DST:
        case SEL_SUBDIVISION_1_NAME_DST:
        case SEL_SUBDIVISION_2_ISO_CODE_DST:
        case SEL_SUBDIVISION_2_NAME_DST:
        case SEL_CITY_NAME_DST:
        case SEL_TIME_ZONE_DST:
        case SEL_REGISTERED_COUNTRY_GEONAME_ID_DST:
        case SEL_REPRESENTED_COUNTRY_GEONAME_ID_DST:
        case SEL_POSTAL_CODE_DST:
        case SEL_AS_ORG_DST:
          switch(sel->field) {
            case SEL_PREFIX:
              line_string = (char *)&opt.line->log_label;
              break;
            case SEL_LOCAL_HOST:
              line_string = (char *)&opt.line->hostname;
              break;
            case SEL_IN_IF:
              line_string = (char *)&opt.line->inif;
              break;
            case SEL_OUT_IF:
              line_string = (char *)&opt.line->outif;
              break;
            case SEL_MAC_SADDR:
              line_string = (char *)&opt.line->mac_saddr_str;
              break;
            case SEL_MAC_DADDR:
              line_string = (char *)&opt.line->mac_daddr_str;
              break;
            case SEL_MAC:
              line_string = (char *)&opt.line->raw_mac;
              break;
            case SEL_CONTINENT_CODE_SRC:
              line_string = (char *)&opt.line->geoip_locations_src->continent_code;
              break;
            case SEL_CONTINENT_NAME_SRC:
              line_string = (char *)&opt.line->geoip_locations_src->continent_name;
              break;
            case SEL_COUNTRY_ISO_CODE_SRC:
              line_string = (char *)&opt.line->geoip_locations_src->country_iso_code;
              break;
            case SEL_COUNTRY_NAME_SRC:
              line_string = (char *)&opt.line->geoip_locations_src->country_name;
              break;
            case SEL_SUBDIVISION_1_ISO_CODE_SRC:
              line_string = (char *)&opt.line->geoip_locations_src->subdivision_1_iso_code;
              break;
            case SEL_SUBDIVISION_1_NAME_SRC:
              line_string = (char *)&opt.line->geoip_locations_src->subdivision_1_name;
              break;
            case SEL_SUBDIVISION_2_ISO_CODE_SRC:
              line_string = (char *)&opt.line->geoip_locations_src->subdivision_2_iso_code;
              break;
            case SEL_SUBDIVISION_2_NAME_SRC:
              line_string = (char *)&opt.line->geoip_locations_src->subdivision_2_name;
              break;
            case SEL_CITY_NAME_SRC:
              line_string = (char *)&opt.line->geoip_locations_src->city_name;
              break;
            case SEL_TIME_ZONE_SRC:
              line_string = (char *)&opt.line->geoip_locations_src->time_zone;
              break;
            case SEL_REGISTERED_COUNTRY_GEONAME_ID_SRC:
              line_string = (char *)&opt.line->geoip_locations_src->registered_country_geoname_id;
              break;
            case SEL_REPRESENTED_COUNTRY_GEONAME_ID_SRC:
              line_string = (char *)&opt.line->geoip_locations_src->represented_country_geoname_id;
              break;
            case SEL_POSTAL_CODE_SRC:
              line_string = (char *)&opt.line->geoip_locations_src->postal_code;
              break;
            case SEL_AS_ORG_SRC:
              line_string = (char *)&opt.line->geoip_asns_src->autonomous_system_organization;
              break;
            case SEL_CONTINENT_CODE_DST:
              line_string = (char *)&opt.line->geoip_locations_dst->continent_code;
              break;
            case SEL_CONTINENT_NAME_DST:
              line_string = (char *)&opt.line->geoip_locations_dst->continent_name;
              break;
            case SEL_COUNTRY_ISO_CODE_DST:
              line_string = (char *)&opt.line->geoip_locations_dst->country_iso_code;
              break;
            case SEL_COUNTRY_NAME_DST:
              line_string = (char *)&opt.line->geoip_locations_dst->country_name;
              break;
            case SEL_SUBDIVISION_1_ISO_CODE_DST:
              line_string = (char *)&opt.line->geoip_locations_dst->subdivision_1_iso_code;
              break;
            case SEL_SUBDIVISION_1_NAME_DST:
              line_string = (char *)&opt.line->geoip_locations_dst->subdivision_1_name;
              break;
            case SEL_SUBDIVISION_2_ISO_CODE_DST:
              line_string = (char *)&opt.line->geoip_locations_dst->subdivision_2_iso_code;
              break;
            case SEL_SUBDIVISION_2_NAME_DST:
              line_string = (char *)&opt.line->geoip_locations_dst->subdivision_2_name;
              break;
            case SEL_CITY_NAME_DST:
              line_string = (char *)&opt.line->geoip_locations_dst->city_name;
              break;
            case SEL_TIME_ZONE_DST:
              line_string = (char *)&opt.line->geoip_locations_dst->time_zone;
              break;
            case SEL_REGISTERED_COUNTRY_GEONAME_ID_DST:
              line_string = (char *)&opt.line->geoip_locations_dst->registered_country_geoname_id;
              break;
            case SEL_REPRESENTED_COUNTRY_GEONAME_ID_DST:
              line_string = (char *)&opt.line->geoip_locations_dst->represented_country_geoname_id;
              break;
            case SEL_POSTAL_CODE_DST:
              line_string = (char *)&opt.line->geoip_locations_dst->postal_code;
              break;
            case SEL_AS_ORG_DST:
              line_string = (char *)&opt.line->geoip_asns_dst->autonomous_system_organization;
              break;
            default:
              break;
          }

          res = regexec(&sel->buf, line_string, 0, NULL, 0);
          if ((res == 0 && sel->invert) || (res == REG_NOMATCH && !sel->invert))
            match = P_MATCH_EXC;
          break; 
        /* End Regex types */
        /* Begin date types */
        case SEL_MIN_DATE:
          if (opt.line->time < (time_t)sel->value)
            match = P_MATCH_EXC;
          break;
        case SEL_MAX_DATE:
          if (opt.line->time > (time_t)sel->value)
            match = P_MATCH_EXC;
          break;
        /* End date types */
        /* TCP option flags */
        case SEL_TCP_FLAGS:
          if (opt.line->protocol == 6 &&
            !(opt.line->flags & TCP_FLAGS_MATCH)) {
            if (sel->value & TCP_OPTS_EXACT) {
              if (opt.line->flags == (sel->value ^ TCP_OPTS_EXACT))
                break;
            } else {
              if (!(opt.line->flags | sel->value) ||
                   (opt.line->flags & sel->value))
                break;
            }
            match = P_MATCH_EXC;
          }
          break;
        /* End TCP options */
        /* Begin IP addresses */
        case SEL_SOURCEHOST:
        case SEL_DESTHOST:
        case SEL_ICMP_GATEWAY:
          switch(sel->field) {
            case SEL_SOURCEHOST:
              host = opt.line->shost;
#if HAVE_INET_NTOP
              host6 = opt.line->shost6;
#endif
              match_value = MATCH_SRC_HOST;
              break;
            case SEL_DESTHOST:
              host = opt.line->dhost;
#if HAVE_INET_NTOP
              host6 = opt.line->dhost6;
#endif
              match_value = MATCH_DST_HOST;
              break;
            case SEL_ICMP_GATEWAY:
              extra_test = opt.line->protocol != 1 || opt.line->icmp_type != 5;
              host = opt.line->icmp_gw;
              match_value = MATCH_ICMP_GATEWAY;
              break;
            default:
              break;
          }
          if (extra_test || (sel->invert && (match_exc & match_value)) || (!sel->invert && (match_inc & match_value)))
            break;
          if (sel->family != opt.line->family) {
            if (!sel->invert) {
              match_exc |= match_value;
              match_inc &= ~match_value;
            } else {
              match_inc |= match_value;
              match_exc &= ~match_value;
            }
            break;
          }
          if (sel->family == AF_INET && opt.line->family == AF_INET) {
            if (((ntohl(host.s_addr) & sel->netmask) != sel->in_addr
                && !(sel->invert))
              || ((ntohl(host.s_addr) & sel->netmask) == sel->in_addr
                && (sel->invert))) { 
              match_exc |= match_value;
              match_inc &= ~match_value;
            } else {
              match_inc |= match_value;
              match_exc &= ~match_value;
            }
#if HAVE_INET_NTOP
          } else if (sel->family == AF_INET6 && opt.line->family == AF_INET6){
            match_inet6 = bitncmp(host6.s6_addr, sel->in6_addr.s6_addr, sel->in6_bits);
            if ((match_inet6 && !sel->invert) || (!match_inet6 && sel->invert)) {
              match_exc |= match_value;
              match_inc &= ~match_value;
            } else {
              match_inc |= match_value;
              match_exc &= ~match_value;
            }
#endif
          }
          break;
        /* End IP addresses */
        default:
          if (opt.verbose >= VERBOSE_ALERT)
              fprintf(stderr, "Skipping unrecognized selection key during parse: %u\n", sel->field);
          break;
      }
      sel = sel->next;
    }

    if(match == P_MATCH_EXC || match_exc != MATCH_MUL_NONE) {
      if (opt.verbose >= VERBOSE_ERROR)
        fprintf(stderr, "e");

      return PARSE_EXCLUDED;
    }

    /* if not excluded, add to data */

#if defined(HAVE_MAXMINDDB_H) && defined(HAVE_DB_H)
    /* mark geoip data used */
    if(opt.line->geoip_locations_src != opt.geoip_locations_init)
      opt.line->geoip_locations_src->using_data = true;
    if(opt.line->geoip_locations_dst != opt.geoip_locations_init)
      opt.line->geoip_locations_dst->using_data = true;
    if(opt.line->geoip_asns_src != opt.geoip_asns_init)
      opt.line->geoip_asns_src->using_data = true;
    if(opt.line->geoip_asns_dst != opt.geoip_asns_init)
      opt.line->geoip_asns_dst->using_data = true;
#endif

    /* opt.line->count is either 0 or 1 */
    opt.orig_count = opt.line->count;
    xstrncpy(opt.line->filename, opt.filename, FILESIZE); 
    opt.line->linenum = linenum;
    build_list();
    if (opt.verbose >= VERBOSE_ERROR) {
      if (repeated)
        fprintf(stderr, "r");
      fprintf(stderr, ".");
    }
  } /* end 'if (retval == PARSE_OK') */
  return retval;
}

/* If parsers specified on command line or in rcfile, use only those */
void select_parsers()
{
  unsigned char i = 0;

  if (opt.format_sel[0] == '\0') {
    return;
  } else {
    opt.format = 0;
    while ((i < SHORTLEN) && (opt.format_sel[i] != '\0')) {
      switch (opt.format_sel[i]) {
      case 'i':
        opt.format = opt.format | PARSER_IPCHAINS;
        break;
      case 'n':
        opt.format = opt.format | PARSER_NETFILTER;
        break;
      case 'f':
        opt.format = opt.format | PARSER_IPFILTER;
        break;
      case 'c':
        opt.format = opt.format | PARSER_CISCO_IOS;
        break;
      case 'p':
        opt.format = opt.format | PARSER_CISCO_PIX;
        break;
      case 'e':
        opt.format = opt.format | PARSER_NETSCREEN;
        break;
      case 'w':
        opt.format = opt.format | PARSER_WIN_XP;
        break;
      case 's':
        opt.format = opt.format | PARSER_SNORT;
        break;
      case 'b':
        opt.format = opt.format | PARSER_IPFW;
        break;

      default:
        fprintf(stderr, "Unknown parser: '%c'.\n", opt.format_sel[i]);
        exit(EXIT_FAILURE);
      }
      i++;
    }
  }
}
