/* $Id: report.c 696 2019-09-20 16:47:49Z bhockney $ */

/* (C) 2004-2019 by Bob Hockney <zeus@ix.netcom.com>    *
 * Based on fwlogwatch written by                       *
 * Boris Wesslowski <Wesslowski@CERT.Uni-Stuttgart.DE>  *
 *                                                      *
 * report generation for wfwl_syslog                    *
 *                                                      *
 * This code is distributed under the terms of GNU GPL  */

#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <pwd.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <syslog.h>
#include <signal.h>

#include <main.h>
#include <parser.h>
#include <compare.h>
#include <utils.h>
#include <database.h>
#include <output.h>

#ifdef HAVE_LIBZ
#include <zlib.h>
#endif

#if defined(HAVE_MAXMINDDB_H) && defined(HAVE_DB_H)
extern DBT key, data;
#endif

/* These are structs have global scope */
extern struct options opt;
extern struct conn_data *first;
extern struct input_file *first_file;
extern struct field_order *fields;
extern struct sort_order *sort;
extern struct fields_used fields_used;
extern struct select_sum select_sum;

/* select based on aggregate criteria */
struct conn_data * select_aggregate(struct conn_data *list) {
  struct conn_data *this, *new=NULL, *add, *old;
  unsigned long int num_rows = 0;
  unsigned long int num_included = 0;
  this = list;

  while (this != NULL) {
    if (!((select_sum.min_count != 0 && this->count < select_sum.min_count) ||
          (select_sum.max_count != 0 && this->count > select_sum.max_count) ||
          (select_sum.max_earliest != 0 && this->earliest > select_sum.max_earliest) ||
          (select_sum.min_latest != 0 && this->latest < select_sum.min_latest))) {
      num_rows++;
      num_included += this->count;
      add = xmalloc(sizeof(struct conn_data));
      *add = *this;
      if (new == NULL) {
        new = add;
        new->next=NULL;
      } else {
        add->next = new;
        new = add;
      }
    }
    old = this;
    this = this->next;
    free(old);
  }

  opt.report_rows = num_rows;
  opt.included_entries = num_included;

  return new;

}

/* PARSE INPUT FILES AND OUTPUT REPORT */
/* This function is the main routine for parsing and outputting reports */
/* It calls other functions as needed */
/* Called without parameters; uses global structs, above */
void generate_report()
{
  char buf[BUFSIZE], *input = NULL, last_file = 0;
  char * ptr = NULL;
  FILE *output = NULL;
  unsigned int j;
  int retval, hit = 0;
  unsigned long linenum = 0;
  unsigned long lines_processed = 0;
  unsigned long lines_matched = 0;
  unsigned long lines_no_hit = 0;
  unsigned long lines_excluded = 0;
  unsigned long lines_wrong_format = 0;
  struct input_file *file;
  struct field_order *flds;
  struct sort_order *srt;
  struct timeval *segtime;
  int64_t hostsupdtime = 0, servicesupdtime = 0, populatecachetime = 0;
  struct geoip_locations *geoip_locations_init;
  struct geoip_asns *geoip_asns_init;
#ifdef HAVE_MAXMINDDB_H
  int status;
  MMDB_s mmdb_loc;
  MMDB_s mmdb_asn;
#ifdef HAVE_DB_H  
  char db_loc_file[L_tmpnam];
  char db_asn_file[L_tmpnam];
  char * dbfile = NULL;
  int tmp, ret;
  uint32_t using_data = 1;
  int x, y;
  DB *dbp;
  DBC *dbcp;
#endif
#endif

  opt.line = xmalloc(sizeof(struct log_line));

#ifdef HAVE_MAXMINDDB_H
  if (strlen(opt.geoip2_file)) {
  status = MMDB_open(opt.geoip2_file, MMDB_MODE_MMAP, &mmdb_loc);

  if (MMDB_SUCCESS != status) {
    opt.use_geoip |= GEOIP_LOC_OFF;
    if(opt.verbose >= VERBOSE_NOTICE) {
      fprintf(stderr, "Can't open mmdb file %s, skipping geoip locations lookup\n  MMDB error is: %s\n", opt.geoip2_file, MMDB_strerror(status));

      if (MMDB_IO_ERROR == status) {
        fprintf(stderr, "  IO error: %s\n", strerror(errno));
      }
    }
  } else { /* success */
    opt.geoip2_locations_handle = &mmdb_loc;
    if (opt.verbose >= VERBOSE_ALERT)
      fprintf(stderr, "Opened mmdb file %s\n", opt.geoip2_file);
    if (opt.verbose >= VERBOSE_NOTICE) {
      if (!(opt.use_geoip & GEOIP_LOC_OFF)) 
        fprintf(stderr, "Resolving geiop locations fields\n");
      else
        fprintf(stderr, "Not resolving geiop locations fields because geoip_php_lookup is set\n");
    }
  }
  } else {
    opt.use_geoip |= GEOIP_LOC_OFF;
    if((opt.use_geoip & (GEOIP_LOC_SRC|GEOIP_LOC_DST)) && opt.verbose >= VERBOSE_INFO)
      fprintf(stderr, "Geoip locations columns, sorts or criteria specified but no locations file available.\n"
                      "  Is the mmdb file installed and configured with geoip2_file set in webfwlog.conf?\n"); 
  }

  if (strlen(opt.geoip2_asn_file)) {
  status = MMDB_open(opt.geoip2_asn_file, MMDB_MODE_MMAP, &mmdb_asn);

  if (MMDB_SUCCESS != status) {
    opt.use_geoip |= GEOIP_ASN_OFF;
    if(opt.verbose >= VERBOSE_NOTICE) {
      fprintf(stderr, "Can't open mmdb file %s, skipping geoip asns lookup\n  MMDB error is: %s\n", opt.geoip2_asn_file, MMDB_strerror(status));

      if (MMDB_IO_ERROR == status) {
        fprintf(stderr, "  IO error: %s\n", strerror(errno));
      }
    }
  } else { /* success */
    opt.geoip2_asns_handle = &mmdb_asn;
    if (opt.verbose >= VERBOSE_ALERT)
      fprintf(stderr, "Opened mmdb file %s\n", opt.geoip2_asn_file);
    if (opt.verbose >= VERBOSE_NOTICE) {
      if (!(opt.use_geoip & GEOIP_ASN_OFF)) 
        fprintf(stderr, "Resolving geiop asn fields\n");
      else
        fprintf(stderr, "Not resolving geiop asn fields because geoip_php_lookup is set\n");
    }
  }
  } else {
    opt.use_geoip |= GEOIP_ASN_OFF;
    if((opt.use_geoip & (GEOIP_ASN_SRC|GEOIP_ASN_DST)) && opt.verbose >= VERBOSE_INFO)
      fprintf(stderr, "Geoip asn columns, sorts or criteria specified but no asn file available.\n"
                      "  Is the mmdb file installed and configured with geoip2_asn_file set in webfwlog.conf?\n"); 
  }
#else
  opt.use_geoip |= GEOIP_LOC_OFF | GEOIP_ASN_OFF;
  if(opt.use_geoip & (GEOIP_LOC_SRC|GEOIP_LOC_DST) && opt.verbose >= VERBOSE_INFO)
    fprintf(stderr, "Geoip locations columns, sorts or criteria specified but no libmaxminddb library available. Is support compiled in?\n"); 
  if(opt.use_geoip & (GEOIP_ASN_SRC|GEOIP_ASN_DST) && opt.verbose >= VERBOSE_INFO)
    fprintf(stderr, "Geoip asn columns, sorts or criteria specified but no libmaxminddb library available. Is support compiled in?\n"); 
#endif

  /* Initialize single struct for geoip data to initialize *
   * geoip fields; only changed if data exists for ip      */
  geoip_locations_init = xmalloc(sizeof(struct geoip_locations));
  geoip_locations_init->continent_code[0] = '\0';
  geoip_locations_init->continent_name[0] = '\0';
  geoip_locations_init->country_iso_code[0] = '\0';
  geoip_locations_init->country_name[0] = '\0';
  geoip_locations_init->subdivision_1_iso_code[0] = '\0';
  geoip_locations_init->subdivision_1_name[0] = '\0';
  geoip_locations_init->subdivision_2_iso_code[0] = '\0';
  geoip_locations_init->subdivision_2_name[0] = '\0';
  geoip_locations_init->city_name[0] = '\0';
  geoip_locations_init->metro_code = 0;
  geoip_locations_init->time_zone[0] = '\0';
  geoip_locations_init->registered_country_geoname_id[0] = '\0';
  geoip_locations_init->represented_country_geoname_id[0] = '\0';
  geoip_locations_init->is_anonymous_proxy = -1;
  geoip_locations_init->is_satellite_provider = -1;
  geoip_locations_init->postal_code[0] = '\0';
  geoip_locations_init->latitude = 0;
  geoip_locations_init->longitude = 0;
  geoip_locations_init->is_in_european_union = -1;
  geoip_locations_init->accuracy_radius = 0;
  geoip_locations_init->using_data = false;
  opt.geoip_locations_init = geoip_locations_init;

  geoip_asns_init = xmalloc(sizeof(struct geoip_asns));
  geoip_asns_init->autonomous_system_number = 0;
  geoip_asns_init->autonomous_system_organization[0] = '\0';
  geoip_asns_init->using_data = false;
  opt.geoip_asns_init = geoip_asns_init;

#if defined(HAVE_MAXMINDDB_H) && defined(HAVE_DB_H)
  for (j=GEOIP_DB_LOC; j < GEOIP_DB_MAX; j++) {
    switch (j) {
      case GEOIP_DB_LOC:
        dbfile = (char *)&db_loc_file;
        break; 
      case GEOIP_DB_ASN:
        dbfile = (char *)&db_asn_file;
        break; 
      default:
        goto err;
        break;
    }

    xstrncpy(dbfile,P_tmpdir,L_tmpnam);
    strncat(dbfile,"/wfwl-XXXXXX",L_tmpnam - strlen(dbfile) - 1);
    tmp = mkstemp(dbfile);

    if (tmp == -1)
      exit(0);
    else
      if (opt.verbose >= VERBOSE_DEBUG)
        fprintf(stderr, "tmp filename %s\n", dbfile);

    close(tmp);
    unlink(dbfile);

    if ((ret = db_create(&dbp, NULL, 0)) != 0) {
      if (opt.verbose >= VERBOSE_INFO)
        fprintf(stderr, "db_create: %s\n", db_strerror(ret));
      goto err;
    }
    if ((ret = dbp->open(dbp, NULL, dbfile, NULL, DB_BTREE, DB_CREATE, 0600)) != 0) {
      if (opt.verbose >= VERBOSE_INFO)
        fprintf(stderr, "db_open: %s\n", db_strerror(ret));
      goto err;
    }

    switch (j) {
      case GEOIP_DB_LOC:
        opt.dbp_locations = dbp;
        if (opt.verbose >= VERBOSE_ALERT)
          fprintf(stderr, "Berkeley locations db opened\n");
        break; 
      case GEOIP_DB_ASN:
        opt.dbp_asns = dbp;
        if (opt.verbose >= VERBOSE_ALERT)
          fprintf(stderr, "Berkeley asns db opened\n");
        break; 
      default:
        goto err;
        break;
    }
  }

  memset(&key, 0, sizeof(key));
  memset(&data, 0, sizeof(data));

  err:

#endif

  if (opt.verbose >= VERBOSE_ERROR) {
    fprintf(stderr, ". = accepted\n");
    fprintf(stderr, "e = excluded\n");
    fprintf(stderr, "r = repeated\n");
    fprintf(stderr, "_ = not packet log entry\n");
  }

  /* get file handle of next file to process, stdin if '-' specified */
  file = first_file;
  while (last_file == 0) {
    if (opt.std_in) {
      if (opt.verbose >= VERBOSE_INFO)
        fprintf(stderr, "Processing input from stdin\n");

#ifdef HAVE_LIBZ
      opt.inputfd = (gzFile) stdin;
#else
      opt.inputfd = stdin;
#endif
    } else {
      input = file->name;
      if (opt.verbose >= VERBOSE_INFO)
        fprintf(stderr, "Processing input file '%s'\n", input);

#ifdef HAVE_LIBZ
      opt.inputfd = gzopen(input, "rb");
#else
      opt.inputfd = fopen(input, "r");
#endif
      if (opt.inputfd == NULL) {
#ifdef HAVE_LIBZ
        fprintf(stderr, "gzopen %s: %s\n", input, strerror(errno));
#else
        fprintf(stderr, "fopen %s: %s\n", input, strerror(errno));
#endif
      }
    }

    /* read first line from input file */
#ifdef HAVE_LIBZ
    if (opt.std_in) {
#endif
      retval = (fgets(buf, BUFSIZE, (FILE *)opt.inputfd) != NULL);
#ifdef HAVE_LIBZ
    } else {
      retval = (gzgets(opt.inputfd, buf, BUFSIZE) != Z_NULL);
    }
#endif

    ptr = strrchr(input, '/');
    if (ptr == NULL)
      ptr = input;
    else
      ptr++;
    xstrncpy(opt.filename, ptr, FILESIZE);
    linenum = 0;

    if(opt.packet) {
      for (j = 1; j < opt.packet; j++) {
#ifdef HAVE_LIBZ
          if (opt.std_in) {
#endif
          retval = (fgets(buf, BUFSIZE, (FILE *)opt.inputfd) != NULL);
#ifdef HAVE_LIBZ
          } else {
            retval = (gzgets(opt.inputfd, buf, BUFSIZE) != Z_NULL);
          }
#endif
      }
    }

    /* process lines until EOF */
    while (retval) {
      lines_processed++;
      hit = parse_line(buf, ++linenum);
      opt.repeated = 0;
      switch (hit) {
        case PARSE_OK:
          lines_matched++;
          opt.matched_entries++;
          opt.repeated = 1;
            break;
        case PARSE_WRONG_FORMAT:
          lines_wrong_format++;
          break;
        case PARSE_EXCLUDED:
          lines_excluded++;
          break;
        case PARSE_NO_HIT:
          lines_no_hit++;
          break;
        default:
          break;
      }

      /* only do one line if packet report */
      if (opt.packet) break;

      /* get next line from input file */
#ifdef HAVE_LIBZ
      if (opt.std_in) {
#endif
        retval = (fgets(buf, BUFSIZE, (FILE *)opt.inputfd) != NULL);
#ifdef HAVE_LIBZ
      } else {
        retval = (gzgets(opt.inputfd, buf, BUFSIZE) != Z_NULL);
      }
#endif
    }

    /* After each input file, print line break following progress indication */
    if (opt.verbose >= VERBOSE_ERROR)
      fprintf(stderr, "\n");

    /* set next input file or indicate done with last_file=1 */
    if(opt.std_in) {
      last_file++;
    } else {
      if (opt.verbose >= VERBOSE_WARNING)
        fprintf(stderr, "Done processing input from '%s'\n", input);

#ifndef HAVE_LIBZ
      retval = fclose(opt.inputfd);
      if (retval == EOF) {
        perror("fclose");
#else
      retval = gzclose(opt.inputfd);
      if (retval != 0) {
        if (retval != Z_ERRNO) {
          fprintf(stderr, "gzclose %s: %s\n", input, gzerror(opt.inputfd, &retval));
        } else {
          perror("gzclose");
        }
#endif
      }

      if (file->next != NULL) {
        file = file->next;
      } else {
        last_file++;
      }
    }
  } /* end while (last_file == 0) */

#ifdef HAVE_MAXMINDDB_H
  if (opt.geoip2_locations_handle !=NULL) {
    MMDB_close(opt.geoip2_locations_handle);
    if(opt.verbose >= VERBOSE_ALERT)
      fprintf(stderr, "Closed mmdb file %s\n", opt.geoip2_file);
  }
  if (opt.geoip2_asns_handle !=NULL) {
    MMDB_close(opt.geoip2_asns_handle);
    if(opt.verbose >= VERBOSE_ALERT)
      fprintf(stderr, "Closed mmdb file %s\n", opt.geoip2_asn_file);
  }
#ifdef HAVE_DB_H
  for (j = GEOIP_DB_LOC; j < GEOIP_DB_MAX; j++) {
    if (j == GEOIP_DB_LOC)
      dbp = opt.dbp_locations;
    else if (j == GEOIP_DB_ASN)
      dbp = opt.dbp_asns;
    else
      continue;
  
    /* Acquire a cursor for the database. */
    ret = dbp->cursor(dbp, NULL, &dbcp, 0);
    if (ret !=0 && opt.verbose >= VERBOSE_INFO)
      fprintf(stderr, "cursor open: %s\n", db_strerror(ret));
  
    /* Re-initialize the key/data pair. */
    memset(&key, 0, sizeof(key));
    memset(&data, 0, sizeof(data));
  
    x = 0;
    y = 0;
  
    while ((ret = dbcp->c_get(dbcp, &key, &data, DB_NEXT)) == 0) {
      y++;
      if (j == GEOIP_DB_LOC)
          using_data = (*(struct geoip_locations **)data.data)->using_data;
      if (j == GEOIP_DB_ASN)
          using_data = (*(struct geoip_asns **)data.data)->using_data;
      
      if (!using_data) {
        free(*(void **)data.data);
        x++;
      }
    }
    if(opt.verbose >= VERBOSE_ALERT) {
      if (j == GEOIP_DB_LOC)
        fprintf(stderr, "Berkeley locations db record count: %u, unused in report: %u\n", y, x);
      else if (j == GEOIP_DB_ASN)
        fprintf(stderr, "Berkeley asns db record count: %u, unused in report: %u\n", y, x);
    }
  
    /* Close the cursor. */
    ret = dbcp->c_close(dbcp);
    if (ret !=0 && opt.verbose >= VERBOSE_INFO)
      fprintf(stderr, "cursor close: %s\n", db_strerror(ret));
  
    ret = dbp->close(dbp, 0);
    if (ret !=0 && opt.verbose >= VERBOSE_INFO)
      fprintf(stderr, "Berkeley db close: %s\n", db_strerror(ret));
  
    if (opt.verbose >= VERBOSE_ALERT) {
      if (j == GEOIP_DB_LOC)
        fprintf(stderr, "Berkeley locations db closed\n");
      else if (j == GEOIP_DB_ASN)
        fprintf(stderr, "Berkeley asns db closed\n");
    }
  }

  if (opt.dbp_locations != NULL)
    unlink((char *)&db_loc_file);
  if (opt.dbp_asns != NULL)
    unlink((char *)&db_asn_file);
#endif
#endif

  free(opt.line);

  if (opt.mode == LOG_SUMMARY)
    first = select_aggregate(first); 

  if (opt.verbose >= VERBOSE_INFO) {
    fprintf(stderr, "Sorting data\n");
  }

  srt = sort;
 
  while (srt != NULL && srt->position != SORT_INV_NO_SORT) {
    opt.sortfield = srt->field;
    opt.sortmode = srt->order;
    first = wfwl_sort(first);
    if(opt.verbose >= VERBOSE_ERROR)
      fprintf(stderr, ".");
    srt = srt->next;
  }

  if (opt.verbose >= VERBOSE_ERROR)
    fprintf(stderr, "\n");

  segtime = xmalloc(sizeof(struct timeval));

  /* resolve host and service names if specified */
  if (opt.resolve_hosts & CACHE_POPULATE || opt.resolve_hosts & CACHE_UPDATE) {
    if (fields_used.dhost_name || fields_used.shost_name) {
      gettimeofday(segtime, NULL);
      if ((opt.resolve_hosts & CACHE_UPDATE))
        hostsupdtime -= segtime->tv_sec;
      else
        populatecachetime -= segtime->tv_sec;
      resolve_hosts();
      gettimeofday(segtime, NULL);
      if (opt.resolve_hosts & CACHE_UPDATE)
        hostsupdtime += segtime->tv_sec;
      else
        populatecachetime += segtime->tv_sec;
    }
    if (fields_used.dst_service || fields_used.src_service) {
      gettimeofday(segtime, NULL);
      if (opt.resolve_hosts & CACHE_UPDATE)
        servicesupdtime -= segtime->tv_sec;
      else
        populatecachetime -= segtime->tv_sec;
      resolve_services();
      gettimeofday(segtime, NULL);
      if (opt.resolve_hosts & CACHE_UPDATE)
        servicesupdtime += segtime->tv_sec;
      else
        populatecachetime += segtime->tv_sec;
    }
  }
  if (opt.resolve_hosts & CACHE_POPULATE_FULL) {
    gettimeofday(segtime, NULL);
    populatecachetime -= segtime->tv_sec;
    populate_cache();
    gettimeofday(segtime, NULL);
    populatecachetime += segtime->tv_sec;
  }

  /* use outfile if specified */
  if (opt.use_out) {
    if (opt.verbose >= VERBOSE_INFO)
      fprintf(stderr, "Writing to output file '%s'\n", opt.outputfile);

    /* redirect stdout to outfile */
    output = freopen(opt.outputfile, "w", stdout);
    if (output == NULL) {
      fprintf(stderr, "freopen %s: %s\n", opt.outputfile, strerror(errno));
      exit(EXIT_FAILURE);
    }
  } else {
    /* otherwise use stdout */
    output = stdout;

    if (opt.verbose >= VERBOSE_ALERT) {
      /* print parse stats */
      fprintf (stderr, "Lines processed:      %7lu\n", lines_processed);
      fprintf (stderr, "Lines matched:        %7lu\n", lines_matched);
      fprintf (stderr, "Matched line repeats: %7lu\n", opt.matched_entries - lines_matched);
      fprintf (stderr, "Lines excluded:       %7lu\n", lines_excluded);
      fprintf (stderr, "Lines wrong format:   %7lu\n", lines_wrong_format);
      fprintf (stderr, "Lines not packet log: %7lu\n", lines_no_hit);
    }

    /* mark end of messages */
    fprintf(output, "%%%%\n");

#ifdef HAVE_MAXMINDDB_H
    if (opt.use_geoip & GEOIP_LOC_OFF || !opt.use_geoip)
#endif
      fprintf(output, "0\n");
#ifdef HAVE_MAXMINDDB_H
    else
      fprintf(output, "1\n");
#endif
#ifdef HAVE_MAXMINDDB_H
    if (opt.use_geoip & GEOIP_ASN_OFF || !opt.use_geoip)
#endif
      fprintf(output, "0\n");
#ifdef HAVE_MAXMINDDB_H
    else
      fprintf(output, "1\n");
#endif

    /* mark end of geoip status */
    fprintf(output, "%%%%\n");
  }

  /* OUTPUT DATA */

  /* output contains file handle for output */

  flds = fields;
  while (flds != NULL) {
    fprintf(output, "%s\t", flds->keyname);
    flds = flds->next;
  }
  fprintf(output, "\n");
  show_list(output);

  fprintf(output, "\n");
  fprintf(output, "%lu\n", opt.report_rows);
  fprintf(output, "%lu\n", opt.matched_entries);
  fprintf(output, "%lu\n", opt.included_entries);
  fprintf(output, "%li\n", (long)hostsupdtime);
  fprintf(output, "%li\n", (long)servicesupdtime);
  fprintf(output, "%li\n", (long)populatecachetime);

  /* clean up */

  free_conn_data();
  free_input_file();

  if (opt.use_out) {
    if (opt.verbose >= VERBOSE_WARNING)
      fprintf(stderr, "Closing '%s'\n", opt.outputfile);

    retval = fclose(output);
    if (retval == EOF) {
      perror("fclose");
    }
  }
}
