/* $Id: utils.c 696 2019-09-20 16:47:49Z bhockney $ */

/* (C) 2004-2019 by Bob Hockney <zeus@ix.netcom.com>    *
 * Based on fwlogwatch written by                       *
 * Boris Wesslowski <Wesslowski@CERT.Uni-Stuttgart.DE>  *
 *                                                      *
 * utility routines for wfwl_syslog                     *
 *                                                      *
 * This code is distributed under the terms of GNU GPL  */

#include <unistd.h>
#include <string.h>
#include <netdb.h>
#include <math.h>

#include <main.h>
#include <utils.h>

struct options opt;
struct input_file *first_file = NULL;
struct conn_data *first = NULL;

/* Allocates memory; returns fatal error if allocation error. *
 * Like malloc, but returns fatal error if unsuccessful       */
void *xmalloc(int size)
{
  void *ptr;

  ptr = malloc(size);
  if (ptr == NULL) {
    fprintf(stderr, "\nMemory allocation error, exiting.\n");
    exit(EXIT_FAILURE);
  }

  return ptr;
}

/* xstrncpy() - similar to strncpy(3) but always terminates string *
 * with '\0' (if n > 0 and dest != NULL),  doesn't do padding.     */
char *xstrncpy(char *dest, const char *src, size_t n)
{
  char *r = dest;

  if((n <= 0) || (dest == NULL)) {
    return dest;
  }
  if(src != NULL) {
    while ((--n != 0) && (*src != '\0')) {
      *dest++ = *src++;
    }
  }
  *dest = '\0';
  return r;
}

/* adds input file to list of files to parse */
void add_input_file(char *name)
{
  struct input_file *file, *ptr;

  if(!strncmp(name, "-", FILESIZE))
    opt.std_in = 1;

  if (opt.std_in) {
    opt.filecount = 0;
    return;
  }

  file = xmalloc(sizeof(struct input_file));
  file->name = xmalloc(FILESIZE);
  file->next = NULL;

  xstrncpy(file->name, name, FILESIZE);

  ptr = first_file;
  if (ptr == NULL) {
    first_file = file;
  } else {
    while (ptr->next != NULL) {
      ptr = ptr->next;
    }
    ptr->next = file;
  }
  opt.filecount++;
}

/* free memory for input files after parsing */
void free_input_file()
{
  struct input_file *file;

  file = first_file;
  while (file != NULL) {
    free(file->name);
    first_file = file;
    file = file->next;
    free(first_file);
  }
  first_file = NULL;
}

/* set initial values for parsed line */
void init_line()
{
  opt.line->filename[0] = '\0';
  opt.line->linenum = 0;
  opt.line->time = 0;
  opt.line->hostname[0] = '\0';
  opt.line->log_label[0] = '\0';
  opt.line->chainlabel[0] = '\0';
  opt.line->branchname[0] = '\0';
  opt.line->interface[0] = '\0';
  opt.line->protocol = 0;
  opt.line->totlen = 0;
  opt.line->family = AF_INET;
  opt.line->shost.s_addr = 0;
#ifdef HAVE_INET_PTON
  inet_pton(AF_INET6, "::", &opt.line->shost6);
#endif
  opt.line->shost_s[0] = '\0';
  opt.line->shostname[0] = '\0';
  opt.line->sport = 65536; /* Sort last if parser or proto doesn't use source port */
  opt.line->sservice[0] = '\0';
  opt.line->dhost.s_addr = 0;
#ifdef HAVE_INET_PTON
  inet_pton(AF_INET6, "::", &opt.line->dhost6);
#endif
  opt.line->dhost_s[0] = '\0';
  opt.line->dhostname[0] = '\0';
  opt.line->dport = 65536; /* Sort last if parser or proto doesn't use destination port */
  opt.line->dservice[0] = '\0';
  opt.line->flags = 0;
  opt.line->count = 0;
  opt.line->oob_protocol = 0;
  opt.line->mac_saddr_str[0] = '\0';
  opt.line->mac_daddr_str[0] = '\0';
  opt.line->raw_mac[0] = '\0';
  opt.line->fwmark = 0;
  opt.line->inif[0] = '\0';
  opt.line->outif[0] = '\0';
  opt.line->tos = 0;
  opt.line->ttl = 0;
  opt.line->ihl = 0;
  opt.line->csum = 0;
  opt.line->ipid = 0;
  opt.line->fragoff = 0;
  opt.line->pay_len = 0;
  opt.line->flowlabel = 0;
  opt.line->tcp_seq = 0;
  opt.line->tcp_ack_seq = 0;
  opt.line->tcp_window = 0;
  opt.line->tcp_urgp = 0;
  opt.line->udp_len = 0;
  opt.line->icmp_type = -1;
  opt.line->icmp_code = -1;
  opt.line->icmp_echoid = 0;
  opt.line->icmp_echoseq = 0;
  opt.line->icmp_gw.s_addr= 0;
  opt.line->icmp_mtu = 0;
  opt.line->ahesp_spi = 0;
  opt.line->geoip_locations_src = opt.geoip_locations_init;
  opt.line->geoip_locations_dst = opt.geoip_locations_init;
  opt.line->geoip_asns_src = opt.geoip_asns_init;
  opt.line->geoip_asns_dst = opt.geoip_asns_init;
}

/* free report data memory when done */
void free_conn_data()
{
  struct conn_data *this;

  this = first;
  while (this != NULL) {
    first = this;
    this = this->next;
    free(first);
  }
  first = NULL;
}

/* Strip trailing newlines from msg */
/* Postresql does this for error messages */
char * strip_nl(char *msg) {
  char *nl;

  nl = strrchr(msg, '\n');
  while (nl != NULL) {
    *nl = '\0';
    nl = strrchr(msg, '\n');
  }

  return msg;
}

/* set opt.line->time = specified time */
void build_time(char *smonth, int day, int hour, int minute, int second)
{
  int month = 0, now, then;
  struct tm *t;
  time_t rr_now;

  rr_now = time(NULL);
  t = localtime(&rr_now);
  now = (int)mktime(t);

  if (strncmp(smonth, "Jan", 3) == 0) { month = 0; }
  else if (strncmp(smonth, "Feb", 3) == 0) { month = 1; }
  else if (strncmp(smonth, "Mar", 3) == 0) { month = 2; }
  else if (strncmp(smonth, "Apr", 3) == 0) { month = 3; }
  else if (strncmp(smonth, "May", 3) == 0) { month = 4; }
  else if (strncmp(smonth, "Jun", 3) == 0) { month = 5; }
  else if (strncmp(smonth, "Jul", 3) == 0) { month = 6; }
  else if (strncmp(smonth, "Aug", 3) == 0) { month = 7; }
  else if (strncmp(smonth, "Sep", 3) == 0) { month = 8; }
  else if (strncmp(smonth, "Oct", 3) == 0) { month = 9; }
  else if (strncmp(smonth, "Nov", 3) == 0) { month = 10; }
  else if (strncmp(smonth, "Dec", 3) == 0) { month = 11; }
  t->tm_mon = month;
  t->tm_mday = day;
  t->tm_hour = hour;
  t->tm_min = minute;
  t->tm_sec = second;
  t->tm_isdst = -1;
  then = (int)mktime(t);
  if (then > now)
    --t->tm_year;

  opt.line->time = mktime(t);
}

/* Returns OK if ip address string ip is valid for address *
 * family.  If address is valid, also sets addr or addr6   *
 * as appropriate. If ip_addr_str is not NULL, the address *
 * is copied there, shortened for AF_INET6 as returned by  *
 * inet_ntop                                               */
unsigned char convert_ip(unsigned short int family, char *ip, struct in_addr *addr, void *addr6, char *ip_addr_str) 
{
#if HAVE_INET_ATON || HAVE_INET_PTON
  int retval;
#endif

  if (family == AF_INET) {
#ifdef HAVE_INET_ATON
  retval = inet_aton(ip, addr);
  if (retval == 0) {
#else
#ifndef INADDR_NONE
#define INADDR_NONE -1
#endif
  addr->s_addr = inet_addr(ip);
  if (addr->s_addr == INADDR_NONE) {
#endif
    if (opt.verbose >= VERBOSE_NOTICE)
      fprintf(stderr, "IP address error: %s\n", ip);
    return IN_ADDR_ERROR;
  }
  if (ip_addr_str != NULL)
    xstrncpy(ip_addr_str, ip, INET6_ADDRSTRLEN);
  return IN_ADDR_OK;
#ifdef HAVE_INET_PTON
  } else if (family == AF_INET6) {
    retval = inet_pton(AF_INET6, ip, addr6);
    if (retval == 0) {
      if (opt.verbose >= VERBOSE_NOTICE)
        fprintf(stderr, "IP address error: %s\n", ip);
      return IN_ADDR_ERROR;
    }
    if (ip_addr_str != NULL)
      inet_ntop(AF_INET6, addr6, ip_addr_str, INET6_ADDRSTRLEN);
    return IN_ADDR_OK;
#endif
  }
  return IN_ADDR_ERROR;
}

/* Compares memory locations l and r for n bits. */
/* Used to compare IPv6 addresses in memory.     */
/* Returns 0 if a match, non-zero otherwise.     */
/* Assumes network byte order.                   */
int bitncmp(void *l, void *r, unsigned int n) {
  unsigned int lb, rb;
  size_t b;
  int x;

  b = n / 8;
  x = memcmp(l, r, b);
  if (x || (n % 8) == 0)
    return x;

  lb = ((const unsigned char *) l)[b];
  rb = ((const unsigned char *) r)[b];
  for (b = n % 8; b > 0; b--) {
    if (((unsigned char)lb & 0x80) != ((unsigned char)rb & 0x80))
      return 1;
    
    lb <<= 1;
    rb <<= 1;
  }

  return 0;
}

/* Zeros bits of location l after n bits, up to len bytes total length   *
 * of l. Used for creating network address of IPv6 address given prefix  *
 * bits. Returns 0 on success. Assumes network byte order.               */
int zero_after_n_bits(void *l, unsigned char n, size_t len) {
  unsigned char b;
  unsigned char *ptr;

  if(n == len * 8)
    return 0;

  ptr = (unsigned char *)l + n/8;
  b = n % 8;
  if (b) {
    *ptr &= ~(unsigned char)(pow(2,(8-b))-1);
     ptr++;
  }

  memset (ptr, 0x00, len - n/8);

  return 0;
}

/* returns host name */
char * resolve_hostname(void *ip, char *type, unsigned long length) {
  struct hostent *host;
  unsigned short int family = AF_INET;
  struct in_addr *in;
#if HAVE_INET_NTOP
  struct in6_addr *in6;
#else
  void *in6;
#endif

  in = xmalloc(sizeof(struct in_addr)); 
#if HAVE_INET_NTOP
  in6 = xmalloc(sizeof(struct in6_addr)); 
#else
  in6 = NULL;
#endif

  /*  If type is "inet", then ip is a string   *
   *  representation of ip address.            *
   *  If type is "binary" then ip is a 16 byte *
   *  binary character string.                 *
   *  Otherwise ip is a struct in_addr.        */        
  if (strncmp(type, "inet", 4) == 0) {
#if HAVE_INET_NTOP
    if (strchr(ip, ':'))
      family = AF_INET6;
    else
#endif
      family = AF_INET;
    if (convert_ip(family, ip, in, in6, NULL) == IN_ADDR_ERROR)
      return "-";
  } else if (strncmp(type, "bin", 3) == 0) {
#if HAVE_INET_NTOP
    if (length == 16) {
      family = AF_INET6;
      memcpy(in6, ip, sizeof(struct in6_addr));
    } else {
#endif
      family = AF_INET;
      memcpy(in, ip, sizeof(struct in_addr));
#if HAVE_INET_NTOP
    }
#endif
  } else
    memcpy(in, ip, sizeof(struct in_addr));

  switch (family) {
    case AF_INET:
      host = gethostbyaddr((void *)in, sizeof(struct in_addr), AF_INET);
      break;
#if HAVE_INET_NTOP
    case AF_INET6:
      host = gethostbyaddr((void *)in6, sizeof(struct in6_addr), AF_INET6);
      break;
#endif
    default:
      host = NULL;
      break;
  }


  if (host != NULL) {
    return ((char *)host->h_name);
  } else {
    return "-";
  }

  free (in);
  free (in6);

}

/* returns proto by name */
char * resolve_protocol(int proto)
{
  struct protoent *protoent;

  protoent = getprotobynumber(proto);
  if (protoent != NULL) {
    return (protoent->p_name);
  } else {
    char *number;
    number = xmalloc(4);
    snprintf(number, 4, "%d", proto);
    return (number);
  }
}

/* returns service name for port and protocol */
char * get_service(unsigned int port, int proto) {
  struct servent *serv;

  serv = getservbyport(htons(port), resolve_protocol(proto));
  if (serv != NULL) {
    return (serv->s_name);
  } else {
    return "-";
  }
}

/* returns port number for service name and protocol */
long int get_port(char *service, int proto) {
  static struct servent *serv;

  serv = getservbyname(service, resolve_protocol(proto));
  if (serv != NULL) {
    return ntohs(serv->s_port);
  } else {
    return -1;
  }
}

/* ICMP type and code tables to resolve and reverse resolve names. */
/* Lables are those used by ipfilter/ipmon to resolve names to numerical */
/*   representations. */
static struct icmp_code icmp_unreach[] = {
  { 0, "net" },
  { 1, "host" },
  { 2, "protocol" },
  { 3, "port" },
  { 4, "needfrag" },
  { 5, "srcfail" },
  { 6, "net_unknown" },
  { 7, "host_unknown" },
  { 8, "isolated" },
  { 9, "net_prohib" },
  {10, "host_prohib" },
  {11, "tosnet" },
  {12, "toshost" },
  {13, "admin_prohibit" }
};

static struct icmp_code icmp_redirect[] = {
  { 0, "net" },
  { 1, "host" },
  { 2, "tosnet" },
  { 3, "toshost" },
};

static struct icmp_code icmp_timeexceed[] = {
  { 0, "transit" },
  { 1, "reassem" }
};

static struct icmp_code icmp_parameter[] = {
  { 0, "errata_pointer" },
  { 1, "optmissing" },
  { 2, "length" },
};

static struct icmp_type icmp_types[] = {
  { 0, "echoreply", NULL, 0 },
  { 3, "unreach", icmp_unreach,
       sizeof(icmp_unreach)/sizeof(struct icmp_code) },
  { 4, "sourcequench", NULL, 0 },
  { 5, "redirect", icmp_redirect,
       sizeof(icmp_redirect)/sizeof(struct icmp_code) },
  { 8, "echo", NULL, 0 },
  { 9, "routeradvert", NULL, 0 },
  {10, "routersolicit", NULL, 0 },
  {11, "timxceed", icmp_timeexceed,
       sizeof(icmp_timeexceed)/sizeof(struct icmp_code) },
  {12, "paramprob", icmp_parameter,
       sizeof(icmp_parameter)/sizeof(struct icmp_code) },
  {13, "timestamp", NULL, 0 },
  {14, "timestampreply", NULL, 0 },
  {15, "inforeq", NULL, 0 },
  {16, "inforeply", NULL, 0 },
  {17, "maskreq", NULL, 0 },
  {18, "maskreply", NULL, 0 },
};

static struct icmp_code icmp6_redirect[] = {
  {  0, "noroute" },
  {  1, "admin" },
  {  2, "neighbour" },
  {  3, "address" },
  {  4, "noport" }
};

static struct icmp_code icmp6_timexceed[] = {
  {  0, "intransit" },
  {  1, "reassem" }
};

static struct icmp_code icmp6_paramprob[] = {
  {  0, "header" },
  {  1, "nextheader" },
  {  2, "option" }
};

static struct icmp_code icmp6_nodeinforequest[] = {
  {  0, "ipv6" },
  {  1, "fqdn" },
  {  2, "ipv4" }
};

static struct icmp_code icmp6_nodeinforeply[] = {
  {  0, "success" },
  {  1, "refused" },
  {  2, "unknown" }
};

static struct icmp_code icmp6_renumber[] = {
  {  0, "command" },
  {  1, "result" },
  {255, "seqnum_reset" }
};

static struct icmp_type icmp6_types[] = {
  {  1, "unreach", icmp6_redirect,
        sizeof(icmp6_redirect)/sizeof(struct icmp_code) },
  {  2, "toobig", NULL, 0 },
  {  3, "timxceed", icmp6_timexceed,
        sizeof(icmp6_timexceed)/sizeof(struct icmp_code) },
  {  4, "paramprob", icmp6_paramprob,
        sizeof(icmp6_paramprob)/sizeof(struct icmp_code) },
  {128, "echo", NULL, 0 },
  {129, "echoreply", NULL, 0 },
  {130, "groupmemberquery", NULL, 0 },
  {131, "groupmemberreport", NULL, 0 },
  {132, "groupmemberterm", NULL, 0 },
  {133, "routersolicit", NULL, 0 },
  {134, "routeradvert", NULL, 0 },
  {135, "neighborsolicit", NULL, 0 },
  {136, "neighboradvert", NULL, 0 },
  {137, "redirect", NULL, 0 },
  {138, "routerrenumber", icmp6_renumber,
        sizeof(icmp6_renumber)/sizeof(struct icmp_code) },
  {139, "nodeinforequest", icmp6_nodeinforequest,
        sizeof(icmp6_nodeinforequest)/sizeof(struct icmp_code) },
  {140, "nodeinforeply", icmp6_nodeinforeply,
        sizeof(icmp6_nodeinforeply)/sizeof(struct icmp_code) },
  {139, "whoareyourequest", NULL, 0 },
  {140, "whoareyoureply", NULL, 0 },
  {139, "fqdnquery", NULL, 0 },
  {140, "fqdnreply", NULL, 0 },
  {200, "mtraceresponse", NULL, 0 },
  {201, "mtracerequest", NULL, 0 }
};

int numtypes = sizeof(icmp_types)/sizeof(struct icmp_type);
int num6types = sizeof(icmp6_types)/sizeof(struct icmp_type);

int get_icmp_values(char *input) {
  int i, j, n_types;
  int type = -1, code = -1;
  char *type_name, *code_name, *proto;
  struct icmp_type *types;

  type_name = xmalloc(SHORTLEN);
  code_name = xmalloc(SHORTLEN);
  proto = xmalloc(SHORTLEN);

  sscanf(input, "%6[icmpv6] %[a-z0-9_]/%[a-z0-9_]", proto, type_name, code_name);
  if (strncmp(proto, "icmpv6", 6) == 0) {
    types = (struct icmp_type *)&icmp6_types;
    n_types = num6types;
  } else {
    types = (struct icmp_type *)&icmp_types;
    n_types = numtypes;
  }
  type = atoi(type_name);
  if (type == 0 && type_name[0] != '0') type = -1;
  for (i = 0; i < n_types; i++) {
    if (types[i].name != NULL &&
       (strncmp(type_name, types[i].name,
        strlen(types[i].name)) == 0 || types[i].type == type)) {
      type = types[i].type;
      code = atoi(code_name);
      if (types[i].codes != NULL) {
        for (j = 0; j < types[i].numcodes; j++) {
          if (strcmp(code_name, types[i].codes[j].name) == 0) {
            code = types[i].codes[j].code;
            break;
          }
        }
      }
      break;
    }
  }

  if (opt.verbose >= VERBOSE_ALERT)
    if (type == -1)
      fprintf(stderr, "Unrecognized Type %s, ignoring\n", type_name);

  if (opt.verbose >= VERBOSE_DEBUG)
    fprintf(stderr, "Type %s is %i; Code %s is %i\n",
            type_name, type, code_name, code);

  if (type != -1) {
    if (code == -1) code = 0;
    opt.line->icmp_type = type;
    opt.line->icmp_code = code;
    return PARSE_OK;
  } else {
    return PARSE_ERROR;
  }

}
